"""
Some basic dicts and field-in-a-dict manipulation helpers.
"""
import collections.abc
from typing import Any, Union, MutableMapping, Mapping, Tuple, List, Text, Iterable, Optional

FieldPath = Tuple[str, ...]
FieldSpec = Union[None, Text, FieldPath, List[str]]

_UNSET = object()


def parse_field(
        field: FieldSpec,
) -> FieldPath:
    """
    Convert any field into a tuple of nested sub-fields.

    Supported notations:

    * ``None`` (for root of a dict).
    * ``"field.subfield"``
    * ``("field", "subfield")``
    * ``["field", "subfield"]``
    """
    if field is None:
        return tuple()
    elif isinstance(field, str):
        return tuple(field.split('.'))
    elif isinstance(field, (list, tuple)):
        return tuple(field)
    else:
        raise ValueError(f"Field must be either a str, or a list/tuple. Got {field!r}")


def resolve(
        d: Mapping,
        field: FieldSpec,
        default: Any = _UNSET,
):
    """
    Retrieve a nested sub-field from a dict.
    """
    path = parse_field(field)
    try:
        result = d
        for key in path:
            result = result[key]
        return result
    except KeyError:
        if default is _UNSET:
            raise
        else:
            return default


def ensure(
        d: MutableMapping,
        field: FieldSpec,
        value: Any,
):
    """
    Force-set a nested sub-field in a dict.
    """
    result = d
    path = parse_field(field)
    if not path:
        raise ValueError("Setting a root of a dict is impossible. Provide the specific fields.")
    for key in path[:-1]:
        try:
            result = result[key]
        except KeyError:
            result = result.setdefault(key, {})
    result[path[-1]] = value


def cherrypick(
        src: Mapping,
        dst: MutableMapping,
        fields: Optional[Iterable[FieldSpec]],
):
    """
    Copy all specified fields between dicts (from src to dst).
    """
    fields = fields if fields is not None else []
    for field in fields:
        try:
            ensure(dst, field, resolve(src, field))
        except KeyError:
            pass  # absent in the source, nothing to merge


def walk(
        objs,
        nested: Optional[Iterable[FieldSpec]] = None,
):
    """
    Iterate over one or many dicts (and sub-dicts recursively).
    """
    if objs is None:
        return
    elif isinstance(objs, collections.abc.Mapping):
        yield objs
        for subfield in (nested if nested is not None else []):
            try:
                yield resolve(objs, parse_field(subfield))
            except KeyError:
                pass
    elif isinstance(objs, collections.abc.Iterable):
        for obj in objs:
            yield from walk(obj, nested=nested)
    else:
        yield objs  # NB: not a mapping, no nested sub-fields.



import collections.abc

from kopf import SimpleRegistry, GlobalRegistry


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn():
    pass


def test_simple_registry_via_iter(mocker):
    cause = mocker.Mock(event=None, diff=None)

    registry = SimpleRegistry()
    iterator = registry.iter_cause_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_simple_registry_via_list(mocker):
    cause = mocker.Mock(event=None, diff=None)

    registry = SimpleRegistry()
    handlers = registry.get_cause_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_simple_registry_with_minimal_signature(mocker):
    cause = mocker.Mock(event=None, diff=None)

    registry = SimpleRegistry()
    registry.register(some_fn)
    handlers = registry.get_cause_handlers(cause)

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn


def test_global_registry_via_iter(mocker, resource):
    cause = mocker.Mock(resource=resource, event=None, diff=None)

    registry = GlobalRegistry()
    iterator = registry.iter_cause_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_global_registry_via_list(mocker, resource):
    cause = mocker.Mock(resource=resource, event=None, diff=None)

    registry = GlobalRegistry()
    handlers = registry.get_cause_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_global_registry_with_minimal_signature(mocker, resource):
    cause = mocker.Mock(resource=resource, event=None, diff=None)

    registry = GlobalRegistry()
    registry.register_cause_handler(resource.group, resource.version, resource.plural, some_fn)
    handlers = registry.get_cause_handlers(cause)

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn




import json

import pytest
import requests

from kopf.clients.events import post_event


async def test_posting(req_mock):
    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    await post_event(obj=obj, type='type', reason='reason', message='message')

    assert req_mock.post.called
    assert req_mock.post.call_count == 1

    data = json.loads(req_mock.post.call_args_list[0][1]['data'])
    assert data['type'] == 'type'
    assert data['reason'] == 'reason'
    assert data['message'] == 'message'
    assert data['source']['component'] == 'kopf'
    assert data['involvedObject']['apiVersion'] == 'group/version'
    assert data['involvedObject']['kind'] == 'kind'
    assert data['involvedObject']['namespace'] == 'ns'
    assert data['involvedObject']['name'] == 'name'
    assert data['involvedObject']['uid'] == 'uid'


async def test_type_is_v1_not_v1beta1(req_mock):
    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    await post_event(obj=obj, type='type', reason='reason', message='message')

    assert req_mock.post.called

    url = req_mock.post.call_args_list[0][1]['url']
    assert 'v1beta1' not in url
    assert '/api/v1/namespaces/ns/events' in url


async def test_api_errors_logged_but_suppressed(req_mock, assert_logs):
    response = requests.Response()
    error = requests.exceptions.HTTPError("boo!", response=response)
    req_mock.post.side_effect = error

    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    await post_event(obj=obj, type='type', reason='reason', message='message')

    assert req_mock.post.called
    assert_logs([
        "Failed to post an event.*boo!",
    ])


async def test_regular_errors_escalate(req_mock):
    error = Exception('boo!')
    req_mock.post.side_effect = error

    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}

    with pytest.raises(Exception) as excinfo:
        await post_event(obj=obj, type='type', reason='reason', message='message')

    assert excinfo.value is error


async def test_message_is_cut_to_max_length(req_mock):
    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    message = 'start' + ('x' * 2048) + 'end'
    await post_event(obj=obj, type='type', reason='reason', message=message)

    data = json.loads(req_mock.post.call_args_list[0][1]['data'])
    assert len(data['message']) <= 1024  # max supported API message length
    assert '...' in data['message']
    assert data['message'].startswith('start')
    assert data['message'].endswith('end')



import kopf


def test_nothing(invoke, login, real_run):
    result = invoke(['run'])
    assert result.exit_code == 0

    registry = kopf.get_default_registry()
    assert len(registry.resources) == 0


def test_one_file(invoke, login, real_run):
    result = invoke(['run', 'handler1.py'])
    assert result.exit_code == 0

    registry = kopf.get_default_registry()
    assert len(registry.resources) == 1
    resource = list(registry.resources)[0]
    handlers = registry._cause_handlers[resource]._handlers
    assert len(handlers) == 1
    assert handlers[0].id == 'create_fn'


def test_two_files(invoke, login, real_run):
    result = invoke(['run', 'handler1.py', 'handler2.py'])
    assert result.exit_code == 0

    registry = kopf.get_default_registry()
    assert len(registry.resources) == 1
    resource = list(registry.resources)[0]
    handlers = registry._cause_handlers[resource]._handlers
    assert len(handlers) == 2
    assert handlers[0].id == 'create_fn'
    assert handlers[1].id == 'update_fn'


def test_one_module(invoke, login, real_run):
    result = invoke(['run', '-m', 'package.module_1'])
    assert result.exit_code == 0

    registry = kopf.get_default_registry()
    assert len(registry.resources) == 1
    resource = list(registry.resources)[0]
    handlers = registry._cause_handlers[resource]._handlers
    assert len(handlers) == 1
    assert handlers[0].id == 'create_fn'


def test_two_modules(invoke, login, real_run):
    result = invoke(['run', '-m', 'package.module_1', '-m', 'package.module_2'])
    assert result.exit_code == 0

    registry = kopf.get_default_registry()
    assert len(registry.resources) == 1
    resource = list(registry.resources)[0]
    handlers = registry._cause_handlers[resource]._handlers
    assert len(handlers) == 2
    assert handlers[0].id == 'create_fn'
    assert handlers[1].id == 'update_fn'


def test_mixed_sources(invoke, login, real_run):
    result = invoke(['run', 'handler1.py', '-m', 'package.module_2'])
    assert result.exit_code == 0

    registry = kopf.get_default_registry()
    assert len(registry.resources) == 1
    resource = list(registry.resources)[0]
    handlers = registry._cause_handlers[resource]._handlers
    assert len(handlers) == 2
    assert handlers[0].id == 'create_fn'
    assert handlers[1].id == 'update_fn'



import logging
from typing import Optional

import pykube
import requests
import urllib3.exceptions

logger = logging.getLogger(__name__)

# Set in login(), consumed in get_pykube_cfg() and all API calls.
_pykube_cfg: Optional[pykube.KubeConfig] = None


class LoginError(Exception):
    """ Raised when the operator cannot login to the API. """


class AccessError(Exception):
    """ Raised when the operator cannot access the cluster API. """


def login(verify=False):
    """
    Login to Kubernetes cluster, locally or remotely.

    Keep the logged in state or config object in the global variables,
    so that it can be available for future calls via the same function call.

    Automatic refresh/reload of the tokens or objects also should be done here.
    """

    # Pykube login is mandatory. If it fails, the framework will not run at all.
    try:
        import pykube
    except ImportError:
        raise  # mandatory
    else:
        login_pykube(verify=verify)

    # We keep the official client library auto-login only because it was
    # an implied behavior before switching to pykube -- to keep it so (implied).
    try:
        import kubernetes
    except ImportError:
        pass  # optional
    else:
        login_client(verify=verify)


def login_pykube(verify=False):
    global _pykube_cfg
    try:
        _pykube_cfg = pykube.KubeConfig.from_service_account()
        logger.debug("Pykube is configured in cluster with service account.")
    except FileNotFoundError:
        try:
            _pykube_cfg = pykube.KubeConfig.from_file()
            logger.debug("Pykube is configured via kubeconfig file.")
        except (pykube.PyKubeError, FileNotFoundError):
            raise LoginError(f"Cannot authenticate pykube neither in-cluster, nor via kubeconfig.")

    if verify:
        verify_pykube()


def login_client(verify=False):
    import kubernetes.client
    try:
        kubernetes.config.load_incluster_config()  # cluster env vars
        logger.debug("Client is configured in cluster with service account.")
    except kubernetes.config.ConfigException as e1:
        try:
            kubernetes.config.load_kube_config()  # developer's config files
            logger.debug("Client is configured via kubeconfig file.")
        except kubernetes.config.ConfigException as e2:
            raise LoginError(f"Cannot authenticate client neither in-cluster, nor via kubeconfig.")

    if verify:
        verify_client()


def verify_pykube():
    """
    Verify if login has succeeded, and the access configuration is still valid.

    All other errors (e.g. 403, 404) are ignored: it means, the host and port
    are configured and are reachable, the authentication token is accepted,
    and the rest are authorization or configuration errors (not a showstopper).
    """
    try:
        api = get_pykube_api()
        rsp = api.get(version="", base="/")
        rsp.raise_for_status()
        api.raise_for_status(rsp)  # replaces requests's HTTPError with its own.
    except requests.exceptions.ConnectionError as e:
        raise AccessError("Cannot connect to the Kubernetes API. "
                          "Please configure the cluster access.")
    except pykube.exceptions.HTTPError as e:
        if e.code == 401:
            raise AccessError("Cannot authenticate to the Kubernetes API. "
                              "Please login or configure the tokens.")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 401:
            raise AccessError("Cannot authenticate to the Kubernetes API. "
                              "Please login or configure the tokens.")


def verify_client():
    """
    Verify if login has succeeded, and the access configuration is still valid.

    All other errors (e.g. 403, 404) are ignored: it means, the host and port
    are configured and are reachable, the authentication token is accepted,
    and the rest are authorization or configuration errors (not a showstopper).
    """
    import kubernetes.client.rest
    try:
        api = kubernetes.client.CoreApi()
        api.get_api_versions()
    except urllib3.exceptions.HTTPError as e:
        raise AccessError("Cannot connect to the Kubernetes API. "
                          "Please configure the cluster access.")
    except kubernetes.client.rest.ApiException as e:
        if e.status == 401:
            raise AccessError("Cannot authenticate to the Kubernetes API. "
                              "Please login or configure the tokens.")


def get_pykube_cfg() -> pykube.KubeConfig:
    if _pykube_cfg is None:
        raise LoginError("Not logged in with PyKube.")
    return _pykube_cfg


# TODO: add some caching, but keep kwargs in mind. Maybe add a key= for purpose/use-place?
def get_pykube_api(timeout=None) -> pykube.HTTPClient:
    kwargs = dict(timeout=timeout)
    return pykube.HTTPClient(get_pykube_cfg(), **kwargs)



import asyncio
import logging

import kopf
from kopf.reactor.handling import custom_object_handler
from kopf.reactor.registries import GlobalRegistry


async def test_nothing_is_called_when_freeze_is_set(mocker, resource, caplog, assert_logs):
    detect_cause = mocker.patch('kopf.reactor.causation.detect_cause')
    handle_cause = mocker.patch('kopf.reactor.handling.handle_cause')
    patch_obj = mocker.patch('kopf.clients.patching.patch_obj')
    asyncio_sleep = mocker.patch('asyncio.sleep')

    # Nothing of these is actually used, but we need to feed something.
    # Except for namespace+name, which are used for the logger prefixes.
    lifecycle = kopf.lifecycles.all_at_once
    registry = GlobalRegistry()
    event = {'object': {'metadata': {'namespace': 'ns1', 'name': 'name1'}}}

    # This is what makes it frozen.
    freeze = asyncio.Event()
    freeze.set()

    caplog.set_level(logging.DEBUG)
    await custom_object_handler(
        lifecycle=lifecycle,
        registry=registry,
        resource=resource,
        event=event,
        freeze=freeze,
        replenished=asyncio.Event(),
        event_queue=asyncio.Queue(),
    )

    assert not detect_cause.called
    assert not handle_cause.called
    assert not patch_obj.called
    assert not asyncio_sleep.called

    assert_logs([
        r"Ignoring the events due to freeze.",
    ])



import asyncio

from kopf.engines.sleeping import sleep_or_wait


async def test_sleep_or_wait_by_delay_reached(timer):
    event = asyncio.Event()
    with timer:
        unslept = await asyncio.wait_for(sleep_or_wait(0.10, event), timeout=1.0)
    assert 0.10 <= timer.seconds < 0.11
    assert unslept is None


async def test_sleep_or_wait_by_event_set(timer):
    event = asyncio.Event()
    asyncio.get_running_loop().call_later(0.07, event.set)
    with timer:
        unslept = await asyncio.wait_for(sleep_or_wait(0.10, event), timeout=1.0)
    assert 0.06 <= timer.seconds <= 0.08
    assert 0.02 <= unslept  <= 0.04


async def test_sleep_or_wait_with_zero_time_and_event_cleared(timer):
    event = asyncio.Event()
    event.clear()
    with timer:
        unslept = await asyncio.wait_for(sleep_or_wait(0, event), timeout=1.0)
    assert timer.seconds <= 0.01
    assert unslept is None


async def test_sleep_or_wait_with_zero_time_and_event_preset(timer):
    event = asyncio.Event()
    event.set()
    with timer:
        unslept = await asyncio.wait_for(sleep_or_wait(0, event), timeout=1.0)
    assert timer.seconds <= 0.01
    assert not unslept  # 0/None; undefined for such case: both goals reached.



import functools
import sys

import click.testing
import pytest

import kopf
from kopf.cli import main

SCRIPT1 = """
import kopf

@kopf.on.create('zalando.org', 'v1', 'kopfexamples')
def create_fn(spec, **_):
    print('Hello from create_fn!')
    print(repr(spec))
"""

SCRIPT2 = """
import kopf

@kopf.on.update('zalando.org', 'v1', 'kopfexamples')
def update_fn(spec, **_):
    print('Hello from create_fn!')
    print(repr(spec))
"""


@pytest.fixture(autouse=True)
def srcdir(tmpdir):
    tmpdir.join('handler1.py').write(SCRIPT1)
    tmpdir.join('handler2.py').write(SCRIPT2)
    pkgdir = tmpdir.mkdir('package')
    pkgdir.join('__init__.py').write('')
    pkgdir.join('module_1.py').write(SCRIPT1)
    pkgdir.join('module_2.py').write(SCRIPT2)

    sys.path.insert(0, str(tmpdir))
    try:
        with tmpdir.as_cwd():
            yield tmpdir
    finally:
        sys.path.remove(str(tmpdir))


@pytest.fixture(autouse=True)
def clean_default_registry():
    registry = kopf.get_default_registry()
    kopf.set_default_registry(kopf.GlobalRegistry())
    try:
        yield
    finally:
        kopf.set_default_registry(registry)


@pytest.fixture(autouse=True)
def clean_modules_cache():
    # Otherwise, the first loaded test-modules remain there forever,
    # preventing 2nd and further tests from passing.
    for key in list(sys.modules.keys()):
        if key.startswith('package'):
            del sys.modules[key]


@pytest.fixture(autouse=True)
def clean_kubernetes_client():
    try:
        import kubernetes
    except ImportError:
        pass  # absent client is already "clean" (or not "dirty" at least).
    else:
        kubernetes.client.configuration.Configuration.set_default(None)


@pytest.fixture()
def runner():
    runner = click.testing.CliRunner()
    return runner


@pytest.fixture()
def invoke(runner):
    return functools.partial(runner.invoke, main)


@pytest.fixture()
def login(mocker):
    mocker.patch('kopf.clients.auth.login_pykube')
    mocker.patch('kopf.clients.auth.login_client')
    mocker.patch('kopf.clients.auth.verify_pykube')
    mocker.patch('kopf.clients.auth.verify_client')


@pytest.fixture()
def preload(mocker):
    return mocker.patch('kopf.utilities.loaders.preload')


@pytest.fixture()
def real_run(mocker):
    return mocker.patch('kopf.reactor.running.run')



import asyncio
import logging

import freezegun
import pytest

import kopf
from kopf.reactor.causation import HANDLER_CAUSES
from kopf.reactor.handling import custom_object_handler


# The timeout is hard-coded in conftest.py:handlers().
# The extrahandlers are needed to prevent the cycle ending and status purging.
@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
@pytest.mark.parametrize('now, ts', [
    ['2099-12-31T23:59:59', '2020-01-01T00:00:00'],
], ids=['slow'])
async def test_timed_out_handler_fails(
        registry, handlers, extrahandlers, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked, now, ts):
    caplog.set_level(logging.DEBUG)
    name1 = f'{cause_type}_fn'

    cause_mock.event = cause_type
    cause_mock.body.update({
        'status': {'kopf': {'progress': {
            'create_fn': {'started': ts},
            'update_fn': {'started': ts},
            'delete_fn': {'started': ts},
            'resume_fn': {'started': ts},
        }}}
    })

    with freezegun.freeze_time(now):
        await custom_object_handler(
            lifecycle=kopf.lifecycles.one_by_one,
            registry=registry,
            resource=resource,
            event={'type': 'irrelevant', 'object': cause_mock.body},
            freeze=asyncio.Event(),
            replenished=asyncio.Event(),
            event_queue=asyncio.Queue(),
        )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called
    assert not handlers.resume_mock.called

    # Progress is reset, as the handler is not going to retry.
    assert not k8s_mocked.sleep_or_wait.called
    assert k8s_mocked.patch_obj.called

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert patch['status']['kopf']['progress'] is not None
    assert patch['status']['kopf']['progress'][name1]['failure']  # evals to true

    assert_logs([
        "Handler .+ has timed out after",
    ])



import pytest

import kopf


@pytest.fixture(autouse=True)
def clear_default_registry():
    registry = kopf.get_default_registry()
    kopf.set_default_registry(kopf.GlobalRegistry())
    try:
        yield
    finally:
        kopf.set_default_registry(registry)




import os.path
import time

import pykube

import kopf.testing


def test_pods_reacted():

    example_py = os.path.join(os.path.dirname(__file__), 'example.py')
    with kopf.testing.KopfRunner(['run', '--standalone', '--verbose', example_py], timeout=60) as runner:
        name = _create_pod()
        time.sleep(5)  # give it some time to react
        _delete_pod(name)
        time.sleep(1)  # give it some time to react

    assert runner.exception is None
    assert runner.exit_code == 0

    assert f'[default/{name}] Creation event:' in runner.stdout
    assert f'[default/{name}] === Pod killing happens in 30s.' in runner.stdout
    assert f'[default/{name}] Deletion event:' in runner.stdout
    assert f'[default/{name}] === Pod killing is cancelled!' in runner.stdout


def _create_pod():
    api = pykube.HTTPClient(pykube.KubeConfig.from_file())
    pod = pykube.Pod(api, {
        'apiVersion': 'v1',
        'kind': 'Pod',
        'metadata': {'generateName': 'kopf-pod-', 'namespace': 'default'},
        'spec': {
            'containers': [{
                'name': 'the-only-one',
                'image': 'busybox',
                'command': ["sh", "-x", "-c", "sleep 1"],
            }]},
    })
    pod.create()
    return pod.name


def _delete_pod(name):
    api = pykube.HTTPClient(pykube.KubeConfig.from_file())
    pod = pykube.Pod.objects(api, namespace='default').get_by_name(name)
    pod.delete()



import asyncio

import pytest

from kopf.engines.posting import event_queue_var, event_queue_loop_var


@pytest.fixture()
def event_queue_loop(event_loop):
    token = event_queue_loop_var.set(event_loop)
    try:
        yield event_loop
    finally:
        event_queue_loop_var.reset(token)


@pytest.fixture()
def event_queue(event_loop):
    queue = asyncio.Queue(loop=event_loop)
    token = event_queue_var.set(queue)
    try:
        yield queue
    finally:
        event_queue_var.reset(token)



"""
The decorators for the event handlers. Usually used as::

    import kopf

    @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
    def creation_handler(**kwargs):
        pass

"""

# TODO: add cluster=True support (different API methods)

from typing import Optional, Union, Tuple, List, Mapping

from kopf.reactor import causation
from kopf.reactor import handling
from kopf.reactor import registries


def resume(
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        timeout: Optional[float] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        labels: Optional[Mapping] = None,
        annotations: Optional[Mapping] = None):
    """ ``@kopf.on.resume()`` handler for the object resuming on operator (re)start. """
    registry = registry if registry is not None else registries.get_default_registry()
    def decorator(fn):
        registry.register_cause_handler(
            group=group, version=version, plural=plural,
            event=None, initial=True, id=id, timeout=timeout,
            fn=fn, labels=labels, annotations=annotations)
        return fn
    return decorator


def create(
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        timeout: Optional[float] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        labels: Optional[Mapping] = None,
        annotations: Optional[Mapping] = None):
    """ ``@kopf.on.create()`` handler for the object creation. """
    registry = registry if registry is not None else registries.get_default_registry()
    def decorator(fn):
        registry.register_cause_handler(
            group=group, version=version, plural=plural,
            event=causation.CREATE, id=id, timeout=timeout,
            fn=fn, labels=labels, annotations=annotations)
        return fn
    return decorator


def update(
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        timeout: Optional[float] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        labels: Optional[Mapping] = None,
        annotations: Optional[Mapping] = None):
    """ ``@kopf.on.update()`` handler for the object update or change. """
    registry = registry if registry is not None else registries.get_default_registry()
    def decorator(fn):
        registry.register_cause_handler(
            group=group, version=version, plural=plural,
            event=causation.UPDATE, id=id, timeout=timeout,
            fn=fn, labels=labels, annotations=annotations)
        return fn
    return decorator


def delete(
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        timeout: Optional[float] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        optional: Optional[bool] = None,
        labels: Optional[Mapping] = None,
        annotations: Optional[Mapping] = None):
    """ ``@kopf.on.delete()`` handler for the object deletion. """
    registry = registry if registry is not None else registries.get_default_registry()
    def decorator(fn):
        registry.register_cause_handler(
            group=group, version=version, plural=plural,
            event=causation.DELETE, id=id, timeout=timeout,
            fn=fn, requires_finalizer=bool(not optional),
            labels=labels, annotations=annotations)
        return fn
    return decorator


def field(
        group: str, version: str, plural: str,
        field: Union[str, List[str], Tuple[str, ...]],
        *,
        id: Optional[str] = None,
        timeout: Optional[float] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        labels: Optional[Mapping] = None,
        annotations: Optional[Mapping] = None):
    """ ``@kopf.on.field()`` handler for the individual field changes. """
    registry = registry if registry is not None else registries.get_default_registry()
    def decorator(fn):
        registry.register_cause_handler(
            group=group, version=version, plural=plural,
            event=None, field=field, id=id, timeout=timeout,
            fn=fn, labels=labels, annotations=annotations)
        return fn
    return decorator


def event(
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        labels: Optional[Mapping] = None,
        annotations: Optional[Mapping] = None):
    """ ``@kopf.on.event()`` handler for the silent spies on the events. """
    registry = registry if registry is not None else registries.get_default_registry()
    def decorator(fn):
        registry.register_event_handler(
            group=group, version=version, plural=plural,
            id=id, fn=fn, labels=labels, annotations=annotations)
        return fn
    return decorator


# TODO: find a better name: `@kopf.on.this` is confusing and does not fully
# TODO: match with the `@kopf.on.{cause}` pattern, where cause is create/update/delete.
def this(
        *,
        id: Optional[str] = None,
        timeout: Optional[float] = None,
        registry: Optional[registries.SimpleRegistry] = None):
    """
    ``@kopf.on.this()`` decorator for the dynamically generated sub-handlers.

    Can be used only inside of the handler function.
    It is efficiently a syntax sugar to look like all other handlers::

        @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
        def create(*, spec, **kwargs):

            for task in spec.get('tasks', []):

                @kopf.on.this(id=f'task_{task}')
                def create_task(*, spec, task=task, **kwargs):

                    pass

    In this example, having spec.tasks set to ``[abc, def]``, this will create
    the following handlers: ``create``, ``create/task_abc``, ``create/task_def``.

    The parent handler is not considered as finished if there are unfinished
    sub-handlers left. Since the sub-handlers will be executed in the regular
    reactor and lifecycle, with multiple low-level events (one per iteration),
    the parent handler will also be executed multiple times, and is expected
    to produce the same (or at least predictable) set of sub-handlers.
    In addition, keep its logic idempotent (not failing on the repeated calls).

    Note: ``task=task`` is needed to freeze the closure variable, so that every
    create function will have its own value, not the latest in the for-cycle.
    """
    registry = registry if registry is not None else handling.subregistry_var.get()
    def decorator(fn):
        registry.register(id=id, fn=fn, timeout=timeout)
        return fn
    return decorator


def register(
        fn,
        *,
        id: Optional[str] = None,
        timeout: Optional[float] = None,
        registry: Optional[registries.SimpleRegistry] = None,
):
    """
    Register a function as a sub-handler of the currently executed handler.

    Example::

        @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
        def create_it(spec, **kwargs):
            for task in spec.get('tasks', []):

                def create_single_task(task=task, **_):
                    pass

                kopf.register(id=task, fn=create_single_task)

    This is efficiently an equivalent for::

        @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
        def create_it(spec, **kwargs):
            for task in spec.get('tasks', []):

                @kopf.on.this(id=task)
                def create_single_task(task=task, **_):
                    pass
    """
    return this(id=id, timeout=timeout, registry=registry)(fn)



import functools
from unittest.mock import Mock

import pytest

from kopf import SimpleRegistry, GlobalRegistry


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn(x=None):
    pass


@pytest.fixture(params=[
    pytest.param(SimpleRegistry, id='in-simple-registry'),
    pytest.param(GlobalRegistry, id='in-global-registry'),
])
def registry(request):
    return request.param()


@pytest.fixture()
def register_fn(registry, resource):
    if isinstance(registry, SimpleRegistry):
        return registry.register
    if isinstance(registry, GlobalRegistry):
        return functools.partial(registry.register_cause_handler, resource.group, resource.version, resource.plural)
    raise Exception(f"Unsupported registry type: {registry}")


@pytest.fixture(params=[
    pytest.param(None, id='without-diff'),
    pytest.param([], id='with-empty-diff'),
])
def cause_no_diff(request, resource):
    body = {'metadata': {'labels': {'somelabel': 'somevalue'}, 'annotations': {'someannotation': 'somevalue'}}}
    return Mock(resource=resource, event='some-event', diff=request.param, body=body)


@pytest.fixture(params=[
    pytest.param([('op', ('some-field',), 'old', 'new')], id='with-field-diff'),
])
def cause_with_diff(resource):
    body = {'metadata': {'labels': {'somelabel': 'somevalue'}, 'annotations': {'someannotation': 'somevalue'}}}
    diff = [('op', ('some-field',), 'old', 'new')]
    return Mock(resource=resource, event='some-event', diff=diff, body=body)


@pytest.fixture(params=[
    pytest.param(None, id='without-diff'),
    pytest.param([], id='with-empty-diff'),
    pytest.param([('op', ('some-field',), 'old', 'new')], id='with-field-diff'),
])
def cause_any_diff(resource, request):
    body = {'metadata': {'labels': {'somelabel': 'somevalue'}, 'annotations': {'someannotation': 'somevalue'}}}
    return Mock(resource=resource, event='some-event', diff=request.param, body=body)


#
# "Catch-all" handlers are those with event == None.
#

def test_catchall_handlers_without_field_found(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event=None, field=None)
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert handlers


def test_catchall_handlers_with_field_found(cause_with_diff, registry, register_fn):
    register_fn(some_fn, event=None, field='some-field')
    handlers = registry.get_cause_handlers(cause_with_diff)
    assert handlers


def test_catchall_handlers_with_field_ignored(cause_no_diff, registry, register_fn):
    register_fn(some_fn, event=None, field='some-field')
    handlers = registry.get_cause_handlers(cause_no_diff)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({'somelabel': 'somevalue'}, id='with-label'),
    pytest.param({'somelabel': 'somevalue', 'otherlabel': 'othervalue'}, id='with-extra-label'),
])
def test_catchall_handlers_with_labels_satisfied(registry, register_fn, resource, labels):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'labels': labels}})
    register_fn(some_fn, event=None, field=None, labels={'somelabel': 'somevalue'})
    handlers = registry.get_cause_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'somelabel': 'othervalue'}, id='with-other-value'),
    pytest.param({'otherlabel': 'othervalue'}, id='with-other-label'),
])
def test_catchall_handlers_with_labels_not_satisfied(registry, register_fn, resource, labels):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'labels': labels}})
    register_fn(some_fn, event=None, field=None, labels={'somelabel': 'somevalue'})
    handlers = registry.get_cause_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({'somelabel': 'somevalue'}, id='with-label'),
    pytest.param({'somelabel': 'othervalue'}, id='with-other-value'),
])
def test_catchall_handlers_with_labels_exist(registry, register_fn, resource, labels):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'labels': labels}})
    register_fn(some_fn, event=None, field=None, labels={'somelabel': None})
    handlers = registry.get_cause_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'otherlabel': 'othervalue'}, id='with-other-label'),
])
def test_catchall_handlers_with_labels_not_exist(registry, register_fn, resource, labels):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'labels': labels}})
    register_fn(some_fn, event=None, field=None, labels={'somelabel': None})
    handlers = registry.get_cause_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'somelabel': 'somevalue'}, id='with-label'),
    pytest.param({'somelabel': 'othervalue'}, id='with-other-value'),
    pytest.param({'otherlabel': 'othervalue'}, id='with-other-label'),
    pytest.param({'somelabel': 'somevalue', 'otherlabel': 'othervalue'}, id='with-extra-label'),
])
def test_catchall_handlers_without_labels(registry, register_fn, resource, labels):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'labels': labels}})
    register_fn(some_fn, event=None, field=None, labels=None)
    handlers = registry.get_cause_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({'someannotation': 'somevalue'}, id='with-annotation'),
    pytest.param({'someannotation': 'somevalue', 'otherannotation': 'othervalue'}, id='with-extra-annotation'),
])
def test_catchall_handlers_with_annotations_satisfied(registry, register_fn, resource, annotations):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'annotations': annotations}})
    register_fn(some_fn, event=None, field=None, annotations={'someannotation': 'somevalue'})
    handlers = registry.get_cause_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'someannotation': 'othervalue'}, id='with-other-value'),
    pytest.param({'otherannotation': 'othervalue'}, id='with-other-annotation'),
])
def test_catchall_handlers_with_annotations_not_satisfied(registry, register_fn, resource, annotations):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'annotations': annotations}})
    register_fn(some_fn, event=None, field=None, annotations={'someannotation': 'somevalue'})
    handlers = registry.get_cause_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({'someannotation': 'somevalue'}, id='with-annotation'),
    pytest.param({'someannotation': 'othervalue'}, id='with-other-value'),
])
def test_catchall_handlers_with_annotations_exist(registry, register_fn, resource, annotations):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'annotations': annotations}})
    register_fn(some_fn, event=None, field=None, annotations={'someannotation': None})
    handlers = registry.get_cause_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'otherannotation': 'othervalue'}, id='with-other-annotation'),
])
def test_catchall_handlers_with_annotations_not_exist(registry, register_fn, resource, annotations):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'annotations': annotations}})
    register_fn(some_fn, event=None, field=None, annotations={'someannotation': None})
    handlers = registry.get_cause_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'someannotation': 'somevalue'}, id='with-annotation'),
    pytest.param({'someannotation': 'othervalue'}, id='with-other-value'),
    pytest.param({'otherannotation': 'othervalue'}, id='with-other-annotation'),
    pytest.param({'someannotation': 'somevalue', 'otherannotation': 'othervalue'}, id='with-extra-annotation'),
])
def test_catchall_handlers_without_annotations(registry, register_fn, resource, annotations):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'annotations': annotations}})
    register_fn(some_fn, event=None, field=None, annotations=None)
    handlers = registry.get_cause_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels, annotations', [
    pytest.param({'somelabel': 'somevalue'}, {'someannotation': 'somevalue'}, id='with-label-annotation'),
    pytest.param({'somelabel': 'somevalue', 'otherlabel': 'othervalue'}, {'someannotation': 'somevalue'}, id='with-extra-label-annotation'),
    pytest.param({'somelabel': 'somevalue'}, {'someannotation': 'somevalue', 'otherannotation': 'othervalue'}, id='with-label-extra-annotation'),
    pytest.param({'somelabel': 'somevalue', 'otherlabel': 'othervalue'}, {'someannotation': 'somevalue', 'otherannotation': 'othervalue'}, id='with-extra-label-extra-annotation'),
])
def test_catchall_handlers_with_labels_and_annotations_satisfied(registry, register_fn, resource, labels, annotations):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'labels': labels, 'annotations': annotations}})
    register_fn(some_fn, event=None, field=None, labels={'somelabel': 'somevalue'}, annotations={'someannotation': 'somevalue'})
    handlers = registry.get_cause_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'somelabel': 'somevalue'}, id='with-label'),
    pytest.param({'somelabel': 'othervalue'}, id='with-other-value'),
    pytest.param({'otherlabel': 'othervalue'}, id='with-other-label'),
    pytest.param({'somelabel': 'somevalue', 'otherlabel': 'othervalue'}, id='with-extra-label'),
])
def test_catchall_handlers_with_labels_and_annotations_not_satisfied(registry, register_fn, resource, labels):
    cause = Mock(resource=resource, event='some-event', diff=None, body={'metadata': {'labels': labels}})
    register_fn(some_fn, event=None, field=None, labels={'somelabel': 'somevalue'}, annotations={'someannotation': 'somevalue'})
    handlers = registry.get_cause_handlers(cause)
    assert not handlers


#
# Relevant handlers are those with event == 'some-event' (but not 'another-event').
# In the per-field handlers, also with field == 'some-field' (not 'another-field').
# In the label filtered handlers, the relevant handlers are those that ask for 'somelabel'.
# In the annotation filtered handlers, the relevant handlers are those that ask for 'someannotation'.
#

def test_relevant_handlers_without_field_found(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='some-event')
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert handlers


def test_relevant_handlers_with_field_found(cause_with_diff, registry, register_fn):
    register_fn(some_fn, event='some-event', field='some-field')
    handlers = registry.get_cause_handlers(cause_with_diff)
    assert handlers


def test_relevant_handlers_with_field_ignored(cause_no_diff, registry, register_fn):
    register_fn(some_fn, event='some-event', field='some-field')
    handlers = registry.get_cause_handlers(cause_no_diff)
    assert not handlers


def test_relevant_handlers_with_labels_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='some-event', labels={'somelabel': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert handlers


def test_relevant_handlers_with_labels_not_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='some-event', labels={'otherlabel': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers


def test_relevant_handlers_with_annotations_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='some-event', annotations={'someannotation': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert handlers


def test_relevant_handlers_with_annotations_not_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='some-event', annotations={'otherannotation': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers


def test_irrelevant_handlers_without_field_ignored(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='another-event')
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers


def test_irrelevant_handlers_with_field_ignored(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='another-event', field='another-field')
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers

def test_irrelevant_handlers_with_labels_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='another-event', labels={'somelabel': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers


def test_irrelevant_handlers_with_labels_not_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='another-event', labels={'otherlabel': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers


def test_irrelevant_handlers_with_annotations_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='another-event', annotations={'someannotation': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers


def test_irrelevant_handlers_with_annotations_not_satisfied(cause_any_diff, registry, register_fn):
    register_fn(some_fn, event='another-event', annotations={'otherannotation': None})
    handlers = registry.get_cause_handlers(cause_any_diff)
    assert not handlers


#
# The handlers must be returned in order of registration,
# even if they are mixed with-/without- * -event/-field handlers.
#

def test_order_persisted_a(cause_with_diff, registry, register_fn):
    register_fn(functools.partial(some_fn, 1), event=None)
    register_fn(functools.partial(some_fn, 2), event='some-event')
    register_fn(functools.partial(some_fn, 3), event='filtered-out-event')
    register_fn(functools.partial(some_fn, 4), event=None, field='filtered-out-field')
    register_fn(functools.partial(some_fn, 5), event=None, field='some-field')

    handlers = registry.get_cause_handlers(cause_with_diff)

    # Order must be preserved -- same as registered.
    assert len(handlers) == 3
    assert handlers[0].event is None
    assert handlers[0].field is None
    assert handlers[1].event == 'some-event'
    assert handlers[1].field is None
    assert handlers[2].event is None
    assert handlers[2].field == ('some-field',)


def test_order_persisted_b(cause_with_diff, registry, register_fn):
    register_fn(functools.partial(some_fn, 1), event=None, field='some-field')
    register_fn(functools.partial(some_fn, 2), event=None, field='filtered-out-field')
    register_fn(functools.partial(some_fn, 3), event='filtered-out-event')
    register_fn(functools.partial(some_fn, 4), event='some-event')
    register_fn(functools.partial(some_fn, 5), event=None)

    handlers = registry.get_cause_handlers(cause_with_diff)

    # Order must be preserved -- same as registered.
    assert len(handlers) == 3
    assert handlers[0].event is None
    assert handlers[0].field == ('some-field',)
    assert handlers[1].event == 'some-event'
    assert handlers[1].field is None
    assert handlers[2].event is None
    assert handlers[2].field is None

#
# Same function should not be returned twice for the same event/cause.
# Only actual for the cases when the event/cause can match multiple handlers.
#

def test_deduplicated(cause_with_diff, registry, register_fn):
    register_fn(some_fn, event=None, id='a')
    register_fn(some_fn, event=None, id='b')

    handlers = registry.get_cause_handlers(cause_with_diff)

    assert len(handlers) == 1
    assert handlers[0].id == 'a'  # the first found one is returned



import asyncio

import pykube
import requests

from kopf import config
from kopf.clients import auth
from kopf.clients import classes


async def patch_obj(*, resource, patch, namespace=None, name=None, body=None):
    """
    Patch a resource of specific kind.

    Either the namespace+name should be specified, or the body,
    which is used only to get namespace+name identifiers.

    Unlike the object listing, the namespaced call is always
    used for the namespaced resources, even if the operator serves
    the whole cluster (i.e. is not namespace-restricted).
    """

    if body is not None and (name is not None or namespace is not None):
        raise TypeError("Either body, or name+namespace can be specified. Got both.")

    namespace = body.get('metadata', {}).get('namespace') if body is not None else namespace
    name = body.get('metadata', {}).get('name') if body is not None else name
    if body is None:
        nskw = {} if namespace is None else dict(namespace=namespace)
        body = {'metadata': {'name': name}}
        body['metadata'].update(nskw)

    api = auth.get_pykube_api()
    cls = classes._make_cls(resource=resource)
    obj = cls(api, body)

    # The handler could delete its own object, so we have nothing to patch. It is okay, ignore.
    try:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(config.WorkersConfig.get_syn_executor(), obj.patch, patch)
    except pykube.ObjectDoesNotExist:
        pass
    except pykube.exceptions.HTTPError as e:
        if e.code == 404:
            pass
        else:
            raise
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 404:
            pass
        else:
            raise



"""
Watching and streaming watch-events.

Kubernetes client's watching streams are synchronous. To make them asynchronous,
we put them into a `concurrent.futures.ThreadPoolExecutor`,
and yield from there asynchronously.

However, async/await coroutines misbehave with `StopIteration` exceptions
raised by the `next` method: see `PEP-479`_.

As a workaround, we replace `StopIteration` with our custom `StopStreaming`
inherited from `RuntimeError` (as suggested by `PEP-479`_),
and re-implement the generators to make them async.

All of this is a workaround for the standard Kubernetes client's limitations.
They would not be needed if the client library were natively asynchronous.

.. _PEP-479: https://www.python.org/dev/peps/pep-0479/
"""

import asyncio
import logging
from typing import Union

import pykube

from kopf import config
from kopf.clients import auth
from kopf.clients import classes
from kopf.clients import fetching
from kopf.reactor import registries

logger = logging.getLogger(__name__)


class WatchingError(Exception):
    """
    Raised when an unexpected error happens in the watch-stream API.
    """


class StopStreaming(RuntimeError):
    """
    Raised when the watch-stream generator ends streaming.
    Replaces `StopIteration`.
    """


def streaming_next(src):
    """
    Same as `next`, but replaces the `StopIteration` with `StopStreaming`.
    """
    try:
        return next(src)
    except StopIteration as e:
        raise StopStreaming(str(e))


async def streaming_aiter(src, loop=None, executor=None):
    """
    Same as `iter`, but asynchronous and stops on `StopStreaming`, not on `StopIteration`.
    """
    loop = loop if loop is not None else asyncio.get_event_loop()
    while True:
        try:
            yield await loop.run_in_executor(executor, streaming_next, src)
        except StopStreaming:
            return


async def infinite_watch(
        resource: registries.Resource,
        namespace: Union[None, str],
):
    """
    Stream the watch-events infinitely.

    This routine is extracted because it is difficult to test infinite loops.
    It is made as simple as possible, and is assumed to work without testing.

    This routine never ends gracefully. If a watcher's stream fails,
    a new one is recreated, and the stream continues.
    It only exits with unrecoverable exceptions.
    """
    while True:
        async for event in streaming_watch(resource=resource, namespace=namespace):
            yield event
        await asyncio.sleep(config.WatchersConfig.watcher_retry_delay)


async def streaming_watch(
        resource: registries.Resource,
        namespace: Union[None, str],
):
    """
    Stream the watch-events from one single API watch-call.
    """

    # First, list the resources regularly, and get the list's resource version.
    # Simulate the events with type "None" event - used in detection of causes.
    rsp = fetching.list_objs(resource=resource, namespace=namespace)
    resource_version = rsp['metadata']['resourceVersion']
    for item in rsp['items']:
        # FIXME: fix in pykube to inject the missing item's fields from the list's metainfo.
        item.setdefault('kind', rsp['kind'][:-4] if rsp['kind'][-4:] == 'List' else rsp['kind'])
        item.setdefault('apiVersion', rsp['apiVersion'])
        yield {'type': None, 'object': item}

    # Then, watch the resources starting from the list's resource version.
    loop = asyncio.get_event_loop()
    stream = watch_objs(resource=resource, namespace=namespace,
                        timeout=config.WatchersConfig.default_stream_timeout,
                        since=resource_version)
    async for event in streaming_aiter(stream, loop=loop):

        # "410 Gone" is for the "resource version too old" error, we must restart watching.
        # The resource versions are lost by k8s after few minutes (as per the official doc).
        # The error occurs when there is nothing happening for few minutes. This is normal.
        if event['type'] == 'ERROR' and event['object']['code'] == 410:
            logger.debug("Restarting the watch-stream for %r", resource)
            break  # out of for-cycle, to the while-true-cycle.

        # Other watch errors should be fatal for the operator.
        if event['type'] == 'ERROR':
            raise WatchingError(f"Error in the watch-stream: {event['object']}")

        # Ensure that the event is something we understand and can handle.
        if event['type'] not in ['ADDED', 'MODIFIED', 'DELETED']:
            logger.warning("Ignoring an unsupported event type: %r", event)
            continue

        # Yield normal events to the consumer.
        yield event


def watch_objs(*, resource, namespace=None, timeout=None, since=None):
    """
    Watch objects of a specific resource type.

    The cluster-scoped call is used in two cases:

    * The resource itself is cluster-scoped, and namespacing makes not sense.
    * The operator serves all namespaces for the namespaced custom resource.

    Otherwise, the namespace-scoped call is used:

    * The resource is namespace-scoped AND operator is namespaced-restricted.
    """

    params = {}
    if timeout is not None:
        params['timeoutSeconds'] = timeout

    api = auth.get_pykube_api(timeout=None)
    cls = classes._make_cls(resource=resource)
    namespace = namespace if issubclass(cls, pykube.objects.NamespacedAPIObject) else None
    lst = cls.objects(api, namespace=pykube.all if namespace is None else namespace)
    src = lst.watch(since=since, params=params)
    return iter({'type': event.type, 'object': event.object.obj} for event in src)



import pytest
import requests

from kopf.clients.fetching import read_obj


@pytest.mark.resource_clustered  # see `req_mock`
def test_when_present_clustered(req_mock, resource):
    result = {}
    req_mock.get.return_value.json.return_value = result

    crd = read_obj(resource=resource, namespace=None, name='name1')
    assert crd is result

    assert req_mock.get.called
    assert req_mock.get.call_count == 1

    url = req_mock.get.call_args_list[0][1]['url']
    assert 'apis/zalando.org/v1/kopfexamples/name1' in url
    assert 'namespaces/' not in url


def test_when_present_namespaced(req_mock, resource):
    result = {}
    req_mock.get.return_value.json.return_value = result

    crd = read_obj(resource=resource, namespace='ns1', name='name1')
    assert crd is result

    assert req_mock.get.called
    assert req_mock.get.call_count == 1

    url = req_mock.get.call_args_list[0][1]['url']
    assert 'apis/zalando.org/v1/namespaces/ns1/kopfexamples/name1' in url


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [403, 404])
def test_when_absent_with_no_default(req_mock, resource, namespace, status):
    response = requests.Response()
    response.status_code = status
    error = requests.exceptions.HTTPError("boo!", response=response)
    req_mock.get.side_effect = error

    with pytest.raises(requests.exceptions.HTTPError) as e:
        read_obj(resource=resource, namespace=namespace, name='name1')
    assert e.value.response.status_code == status


@pytest.mark.parametrize('default', [None, object()], ids=['none', 'object'])
@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [403, 404])
def test_when_absent_with_default(req_mock, resource, namespace, default, status):
    response = requests.Response()
    response.status_code = status
    error = requests.exceptions.HTTPError("boo!", response=response)
    req_mock.get.side_effect = error

    crd = read_obj(resource=resource, namespace=namespace, name='name1', default=default)
    assert crd is default


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [400, 401, 500, 666])
def test_raises_api_error_despite_default(req_mock, resource, namespace, status):
    response = requests.Response()
    response.status_code = status
    error = requests.exceptions.HTTPError("boo!", response=response)
    req_mock.get.side_effect = error

    with pytest.raises(requests.exceptions.HTTPError) as e:
        read_obj(resource=resource, namespace=namespace, name='name1', default=object())
    assert e.value.response.status_code == status



import copy
import datetime
from unittest.mock import Mock

import freezegun
import pytest

from kopf.structs.status import (
    is_started,
    is_sleeping,
    is_awakened,
    is_finished,
    get_start_time,
    get_awake_time,
    get_retry_count,
    set_start_time,
    set_awake_time,
    set_retry_time,
    store_failure,
    store_success,
    store_result,
    purge_progress,
)

# Timestamps: time zero (0), before (B), after (A), and time zero+1s (1).
TSB = datetime.datetime(2020, 12, 31, 23, 59, 59, 000000)
TS0 = datetime.datetime(2020, 12, 31, 23, 59, 59, 123456)
TS1 = datetime.datetime(2021,  1,  1, 00, 00, 00, 123456)
TSA = datetime.datetime(2020, 12, 31, 23, 59, 59, 999999)
TSB_ISO = '2020-12-31T23:59:59.000000'
TS0_ISO = '2020-12-31T23:59:59.123456'
TS1_ISO = '2021-01-01T00:00:00.123456'
TSA_ISO = '2020-12-31T23:59:59.999999'


@pytest.fixture()
def handler():
    return Mock(id='some-id', spec_set=['id'])


@pytest.mark.parametrize('expected, body', [
    (False, {}),
    (False, {'status': {}}),
    (False, {'status': {'kopf': {}}}),
    (False, {'status': {'kopf': {'progress': {}}}}),
    (False, {'status': {'kopf': {'progress': {'etc-id': {}}}}}),
    (True , {'status': {'kopf': {'progress': {'some-id': {}}}}}),
])
def test_is_started(handler, expected, body):
    origbody = copy.deepcopy(body)
    result = is_started(body=body, handler=handler)
    assert result == expected
    assert body == origbody  # not modified


@pytest.mark.parametrize('body', [
    {},
    {'status': {}},
    {'status': {'kopf': {}}},
    {'status': {'kopf': {'progress': {}}}},
    {'status': {'kopf': {'progress': {'some-id': {}}}}},
])
def test_is_finished_with_partial_status_remains_readonly(handler, body):
    origbody = copy.deepcopy(body)
    result = is_finished(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result
    assert body == origbody  # not modified


@pytest.mark.parametrize('finish_value', [None, False, 'bad'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
def test_is_finished_when_not_finished(handler, finish_field, finish_value):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_finished(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result


@pytest.mark.parametrize('finish_value', [True, 'good'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
def test_is_finished_when_finished(handler, finish_field, finish_value):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_finished(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert result


@pytest.mark.parametrize('body', [
    {},
    {'status': {}},
    {'status': {'kopf': {}}},
    {'status': {'kopf': {'progress': {}}}},
    {'status': {'kopf': {'progress': {'some-id': {}}}}},
])
def test_is_sleeping_with_partial_status_remains_readonly(handler, body):
    origbody = copy.deepcopy(body)
    result = is_finished(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result
    assert body == origbody  # not modified


@pytest.mark.parametrize('finish_value', [True, 'good'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({}, id='delayed-empty'),
    pytest.param({'delayed': None}, id='delayed-none'),
    pytest.param({'delayed': TSB_ISO}, id='delayed-before'),
    pytest.param({'delayed': TS0_ISO}, id='delayed-exact'),
    pytest.param({'delayed': TS1_ISO}, id='delayed-onesec'),
    pytest.param({'delayed': TSA_ISO}, id='delayed-after'),
])
@freezegun.freeze_time(TS0)
def test_is_sleeping_when_finished_regardless_of_delay(
        handler, finish_field, finish_value, delayed_body):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_sleeping(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result


@pytest.mark.parametrize('finish_value', [None, False, 'bad'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({}, id='delayed-empty'),
    pytest.param({'delayed': None}, id='delayed-none'),
])
@freezegun.freeze_time(TS0)
def test_is_sleeping_when_not_finished_and_not_delayed(
        handler, delayed_body, finish_field, finish_value):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_sleeping(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result


@pytest.mark.parametrize('finish_value', [None, False, 'bad'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({'delayed': TSB_ISO}, id='delayed-before'),
    pytest.param({'delayed': TS0_ISO}, id='delayed-exact'),
])
@freezegun.freeze_time(TS0)
def test_is_sleeping_when_not_finished_and_delayed_until_before_now(
        handler, finish_field, finish_value, delayed_body):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_sleeping(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result


@pytest.mark.parametrize('finish_value', [None, False, 'bad'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({'delayed': TS1_ISO}, id='delayed-onesec'),
    pytest.param({'delayed': TSA_ISO}, id='delayed-after'),
])
@freezegun.freeze_time(TS0)
def test_is_sleeping_when_not_finished_and_delayed_until_after_now(
        handler, finish_field, finish_value, delayed_body):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_sleeping(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert result


@pytest.mark.parametrize('body', [
    {},
    {'status': {}},
    {'status': {'kopf': {}}},
    {'status': {'kopf': {'progress': {}}}},
    {'status': {'kopf': {'progress': {'some-id': {}}}}},
])
def test_is_awakened_with_partial_status_remains_readonly(handler, body):
    origbody = copy.deepcopy(body)
    result = is_awakened(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert result
    assert body == origbody  # not modified


@pytest.mark.parametrize('finish_value', [True, 'good'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({}, id='delayed-empty'),
    pytest.param({'delayed': None}, id='delayed-none'),
    pytest.param({'delayed': TSB_ISO}, id='delayed-before'),
    pytest.param({'delayed': TS0_ISO}, id='delayed-exact'),
    pytest.param({'delayed': TS1_ISO}, id='delayed-onesec'),
    pytest.param({'delayed': TSA_ISO}, id='delayed-after'),
])
@freezegun.freeze_time(TS0)
def test_is_awakened_when_finished_regardless_of_delay(
        handler, finish_field, finish_value, delayed_body):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_awakened(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result


@pytest.mark.parametrize('finish_value', [None, False, 'bad'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({}, id='delayed-empty'),
    pytest.param({'delayed': None}, id='delayed-none'),
])
@freezegun.freeze_time(TS0)
def test_is_awakened_when_not_finished_and_not_delayed(
        handler, delayed_body, finish_field, finish_value):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_awakened(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert result


@pytest.mark.parametrize('finish_value', [None, False, 'bad'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({'delayed': TSB_ISO}, id='delayed-before'),
    pytest.param({'delayed': TS0_ISO}, id='delayed-exact'),
])
@freezegun.freeze_time(TS0)
def test_is_awakened_when_not_finished_and_delayed_until_before_now(
        handler, finish_field, finish_value, delayed_body):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_awakened(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert result


@pytest.mark.parametrize('finish_value', [None, False, 'bad'])
@pytest.mark.parametrize('finish_field', ['failure', 'success'])
@pytest.mark.parametrize('delayed_body', [
    pytest.param({'delayed': TS1_ISO}, id='delayed-onesec'),
    pytest.param({'delayed': TSA_ISO}, id='delayed-after'),
])
@freezegun.freeze_time(TS0)
def test_is_awakened_when_not_finished_and_delayed_until_after_now(
        handler, finish_field, finish_value, delayed_body):
    body = {'status': {'kopf': {'progress': {'some-id': {}}}}}
    body['status']['kopf']['progress']['some-id'].update(delayed_body)
    body['status']['kopf']['progress']['some-id'][finish_field] = finish_value
    result = is_awakened(body=body, digest='good', handler=handler)
    assert isinstance(result, bool)
    assert not result


@pytest.mark.parametrize('expected, body', [
    (None, {}),
    (None, {'status': {}}),
    (None, {'status': {'kopf': {}}}),
    (None, {'status': {'kopf': {'progress': {}}}}),
    (None, {'status': {'kopf': {'progress': {'some-id': {}}}}}),
    (None, {'status': {'kopf': {'progress': {'some-id': {'delayed': None}}}}}),
    (TS0, {'status': {'kopf': {'progress': {'some-id': {'delayed': TS0_ISO}}}}}),
])
def test_get_awake_time(handler, expected, body):
    origbody = copy.deepcopy(body)
    result = get_awake_time(body=body, handler=handler)
    assert result == expected
    assert body == origbody  # not modified


@pytest.mark.parametrize('expected, body, patch', [
    (None, {}, {}),
    (None, {'status': {}}, {}),
    (None, {'status': {'kopf': {}}}, {}),
    (None, {'status': {'kopf': {'progress': {}}}}, {}),
    (None, {'status': {'kopf': {'progress': {'some-id': {}}}}}, {}),
    (None, {'status': {'kopf': {'progress': {'some-id': {'started': None}}}}}, {}),
    (TS0, {'status': {'kopf': {'progress': {'some-id': {'started': TS0_ISO}}}}}, {}),
    (None, {}, {'status': {}}),
    (None, {}, {'status': {'kopf': {}}}),
    (None, {}, {'status': {'kopf': {'progress': {}}}}),
    (None, {}, {'status': {'kopf': {'progress': {'some-id': {}}}}}),
    (None, {}, {'status': {'kopf': {'progress': {'some-id': {'started': None}}}}}),
    (TS0, {}, {'status': {'kopf': {'progress': {'some-id': {'started': TS0_ISO}}}}}),
    (TSB,  # the patch has priority
     {'status': {'kopf': {'progress': {'some-id': {'started': TSA_ISO}}}}},
     {'status': {'kopf': {'progress': {'some-id': {'started': TSB_ISO}}}}}),
])
def test_get_start_time(handler, expected, body, patch):
    origbody = copy.deepcopy(body)
    origpatch = copy.deepcopy(patch)
    result = get_start_time(body=body, patch=patch, handler=handler)
    assert result == expected
    assert body == origbody  # not modified
    assert patch == origpatch  # not modified


@pytest.mark.parametrize('expected, body', [
    (0, {}),
    (0, {'status': {}}),
    (0, {'status': {'kopf': {'progress': {}}}}),
    (0, {'status': {'kopf': {'progress': {'some-id': {}}}}}),
    (0, {'status': {'kopf': {'progress': {'some-id': {'retries': None}}}}}),
    (6, {'status': {'kopf': {'progress': {'some-id': {'retries': 6}}}}}),
])
def test_get_retry_count(handler, expected, body):
    origbody = copy.deepcopy(body)
    result = get_retry_count(body=body, handler=handler)
    assert result == expected
    assert body == origbody  # not modified


@pytest.mark.parametrize('body, expected', [
    ({}, {'status': {'kopf': {'progress': {'some-id': {'started': TS0_ISO}}}}}),
])
@freezegun.freeze_time(TS0)
def test_set_start_time(handler, expected, body):
    origbody = copy.deepcopy(body)
    patch = {}
    set_start_time(body=body, patch=patch, handler=handler)
    assert patch == expected
    assert body == origbody  # not modified


@pytest.mark.parametrize('body, delay, expected', [
    ({}, None, {'status': {'kopf': {'progress': {'some-id': {'delayed': None}}}}}),
    ({}, 0, {'status': {'kopf': {'progress': {'some-id': {'delayed': TS0_ISO}}}}}),
    ({}, 1, {'status': {'kopf': {'progress': {'some-id': {'delayed': TS1_ISO}}}}}),
])
@freezegun.freeze_time(TS0)
def test_set_awake_time(handler, expected, body, delay):
    origbody = copy.deepcopy(body)
    patch = {}
    set_awake_time(body=body, patch=patch, handler=handler, delay=delay)
    assert patch == expected
    assert body == origbody  # not modified


@pytest.mark.parametrize('body, delay, expected', [
    ({}, None,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 1, 'delayed': None}}}}}),
    ({}, 0,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 1, 'delayed': TS0_ISO}}}}}),
    ({}, 1,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 1, 'delayed': TS1_ISO}}}}}),

    ({'status': {'kopf': {'progress': {'some-id': {'retries': None}}}}}, None,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 1, 'delayed': None}}}}}),
    ({'status': {'kopf': {'progress': {'some-id': {'retries': None}}}}}, 0,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 1, 'delayed': TS0_ISO}}}}}),
    ({'status': {'kopf': {'progress': {'some-id': {'retries': None}}}}}, 1,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 1, 'delayed': TS1_ISO}}}}}),

    ({'status': {'kopf': {'progress': {'some-id': {'retries': 5}}}}}, None,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 6, 'delayed': None}}}}}),
    ({'status': {'kopf': {'progress': {'some-id': {'retries': 5}}}}}, 0,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 6, 'delayed': TS0_ISO}}}}}),
    ({'status': {'kopf': {'progress': {'some-id': {'retries': 5}}}}}, 1,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 6, 'delayed': TS1_ISO}}}}}),
])
@freezegun.freeze_time(TS0)
def test_set_retry_time(handler, expected, body, delay):
    origbody = copy.deepcopy(body)
    patch = {}
    set_retry_time(body=body, patch=patch, handler=handler, delay=delay)
    assert patch == expected
    assert body == origbody  # not modified


@pytest.mark.parametrize('body, expected', [
    ({},
     {'status': {'kopf': {'progress': {'some-id': {'stopped': TS0_ISO,
                                                   'failure': 'digest',
                                                   'retries': 1,
                                                   'message': 'some-error'}}}}}),

    ({'status': {'kopf': {'progress': {'some-id': {'retries': 5}}}}},
     {'status': {'kopf': {'progress': {'some-id': {'stopped': TS0_ISO,
                                                   'failure': 'digest',
                                                   'retries': 6,
                                                   'message': 'some-error'}}}}}),
])
@freezegun.freeze_time(TS0)
def test_store_failure(handler, expected, body):
    origbody = copy.deepcopy(body)
    patch = {}
    store_failure(body=body, patch=patch, digest='digest',
                  handler=handler, exc=Exception("some-error"))
    assert patch == expected
    assert body == origbody  # not modified


@pytest.mark.parametrize('result, body, expected', [

    # With no result, it updates only the progress.
    (None,
     {},
     {'status': {'kopf': {'progress': {'some-id': {'stopped': TS0_ISO,
                                                   'success': 'digest',
                                                   'retries': 1,
                                                   'message': None}}}}}),
    (None,
     {'status': {'kopf': {'progress': {'some-id': {'retries': 5}}}}},
     {'status': {'kopf': {'progress': {'some-id': {'stopped': TS0_ISO,
                                                   'success': 'digest',
                                                   'retries': 6,
                                                   'message': None}}}}}),

    # With the result, it updates also the status.
    ({'field': 'value'},
     {},
     {'status': {'kopf': {'progress': {'some-id': {'stopped': TS0_ISO,
                                                   'success': 'digest',
                                                   'retries': 1,
                                                   'message': None}}},
                 'some-id': {'field': 'value'}}}),
    ({'field': 'value'},
     {'status': {'kopf': {'progress': {'some-id': {'retries': 5}}}}},
     {'status': {'kopf': {'progress': {'some-id': {'stopped': TS0_ISO,
                                                   'success': 'digest',
                                                   'retries': 6,
                                                   'message': None}}},
                 'some-id': {'field': 'value'}}}),
])
@freezegun.freeze_time(TS0)
def test_store_success(handler, expected, body, result):
    origbody = copy.deepcopy(body)
    patch = {}
    store_success(body=body, patch=patch, digest='digest',
                  handler=handler, result=result)
    assert patch == expected
    assert body == origbody  # not modified



@pytest.mark.parametrize('result, expected', [
    (None,
     {}),
    ({'field': 'value'},
     {'status': {'some-id': {'field': 'value'}}}),
    ('string',
     {'status': {'some-id': 'string'}}),
])
def test_store_result(handler, expected, result):
    patch = {}
    store_result(patch=patch, handler=handler, result=result)
    assert patch == expected


@pytest.mark.parametrize('body', [
    ({}),
    ({'status': {'kopf': {'progress': {'some-id': {'retries': 5}}}}}),
])
def test_purge_progress(body):
    origbody = copy.deepcopy(body)
    patch = {}
    purge_progress(body=body, patch=patch)
    assert patch == {'status': {'kopf': {'progress': None}}}
    assert body == origbody  # not modified



import pytest

from kopf.reactor.registries import Handler


def test_no_args():
    with pytest.raises(TypeError):
        Handler()


def test_all_args(mocker):
    fn = mocker.Mock()
    id = mocker.Mock()
    event = mocker.Mock()
    field = mocker.Mock()
    timeout = mocker.Mock()
    initial = mocker.Mock()
    labels = mocker.Mock()
    annotations = mocker.Mock()
    handler = Handler(
        fn=fn,
        id=id,
        event=event,
        field=field,
        timeout=timeout,
        initial=initial,
        labels=labels,
        annotations=annotations,
    )
    assert handler.fn is fn
    assert handler.id is id
    assert handler.event is event
    assert handler.field is field
    assert handler.timeout is timeout
    assert handler.initial is initial
    assert handler.labels is labels
    assert handler.annotations is annotations



import pytest

from kopf.reactor.registries import Resource


def test_no_args():
    with pytest.raises(TypeError):
        Resource()


def test_all_args(mocker):
    group = mocker.Mock()
    version = mocker.Mock()
    plural = mocker.Mock()
    resource = Resource(
        group=group,
        version=version,
        plural=plural,
    )
    assert resource.group is group
    assert resource.version is version
    assert resource.plural is plural



import asyncio
import json
import logging

import freezegun
import pytest

import kopf
from kopf.reactor.causation import HANDLER_CAUSES, CREATE, UPDATE, DELETE, RESUME
from kopf.reactor.handling import TemporaryError
from kopf.reactor.handling import WAITING_KEEPALIVE_INTERVAL
from kopf.reactor.handling import custom_object_handler
from kopf.structs.finalizers import FINALIZER


@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
@pytest.mark.parametrize('now, ts, delay', [
    ['2020-01-01T00:00:00', '2020-01-01T00:04:56.789000', 4 * 60 + 56.789],
], ids=['fast'])
async def test_delayed_handlers_progress(
        registry, handlers, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked, now, ts, delay):
    caplog.set_level(logging.DEBUG)

    handlers.create_mock.side_effect = TemporaryError("oops", delay=delay)
    handlers.update_mock.side_effect = TemporaryError("oops", delay=delay)
    handlers.delete_mock.side_effect = TemporaryError("oops", delay=delay)
    handlers.resume_mock.side_effect = TemporaryError("oops", delay=delay)

    cause_mock.event = cause_type

    with freezegun.freeze_time(now):
        await custom_object_handler(
            lifecycle=kopf.lifecycles.all_at_once,
            registry=registry,
            resource=resource,
            event={'type': 'irrelevant', 'object': cause_mock.body},
            freeze=asyncio.Event(),
            replenished=asyncio.Event(),
            event_queue=asyncio.Queue(),
        )

    assert handlers.create_mock.call_count == (1 if cause_type == CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == RESUME else 0)

    assert not k8s_mocked.sleep_or_wait.called
    assert k8s_mocked.patch_obj.called

    fname = f'{cause_type}_fn'
    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert patch['status']['kopf']['progress'][fname]['delayed'] == ts

    assert_logs([
        "Invoking handler .+",
        "Handler .+ failed temporarily: oops",
    ])


@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
@pytest.mark.parametrize('now, ts, delay', [
    ['2020-01-01T00:00:00', '2020-01-01T00:04:56.789000', 4 * 60 + 56.789],
    ['2020-01-01T00:00:00', '2099-12-31T23:59:59.000000', WAITING_KEEPALIVE_INTERVAL],
], ids=['fast', 'slow'])
async def test_delayed_handlers_sleep(
        registry, handlers, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked, now, ts, delay):
    caplog.set_level(logging.DEBUG)

    cause_mock.event = cause_type
    cause_mock.body.update({
        'status': {'kopf': {
            'frozen-state': json.dumps({'spec': {}}),  # to prevent re-adding it
            'progress': {
                'create_fn': {'delayed': ts},
                'update_fn': {'delayed': ts},
                'delete_fn': {'delayed': ts},
                'resume_fn': {'delayed': ts},
            },
        }}
    })
    # make sure the finalizer is added since there are mandatory deletion handlers
    cause_mock.body.setdefault('metadata', {})['finalizers'] = [FINALIZER]

    with freezegun.freeze_time(now):
        await custom_object_handler(
            lifecycle=kopf.lifecycles.all_at_once,
            registry=registry,
            resource=resource,
            event={'type': 'irrelevant', 'object': cause_mock.body},
            freeze=asyncio.Event(),
            replenished=asyncio.Event(),
            event_queue=asyncio.Queue(),
        )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called
    assert not handlers.resume_mock.called

    # The dummy patch is needed to trigger the further changes. The value is irrelevant.
    assert k8s_mocked.patch_obj.called
    assert k8s_mocked.patch_obj.call_args_list[-1][1]['patch']  # not empty, maybe with ['dummy']

    # The duration of sleep should be as expected.
    assert k8s_mocked.sleep_or_wait.called
    assert k8s_mocked.sleep_or_wait.call_args_list[0][0][0] == delay

    assert_logs([
        r"Sleeping for [\d\.]+ seconds",
    ])



"""
Invoking the callbacks, including the args/kwargs preparation.

Both sync & async functions are supported, so as their partials.
Also, decorated wrappers and lambdas are recognized.
All of this goes via the same invocation logic and protocol.
"""
import asyncio
import contextvars
import functools
from typing import Callable

from kopf import config


async def invoke(
        fn: Callable,
        *args,
        **kwargs):
    """
    Invoke a single function, but safely for the main asyncio process.

    Used both for the handler functions and for the lifecycle callbacks.

    A full set of the arguments is provided, expanding the cause to some easily
    usable aliases. The function is expected to accept ``**kwargs`` for the args
    that it does not use -- for forward compatibility with the new features.

    The synchronous methods are executed in the executor (threads or processes),
    thus making it non-blocking for the main event loop of the operator.
    See: https://pymotw.com/3/asyncio/executors.html
    """

    # Add aliases for the kwargs, directly linked to the body, or to the assumed defaults.
    if 'event' in kwargs:
        event = kwargs.get('event')
        kwargs.update(
            type=event['type'],
            body=event['object'],
            spec=event['object'].setdefault('spec', {}),
            meta=event['object'].setdefault('metadata', {}),
            status=event['object'].setdefault('status', {}),
            uid=event['object'].get('metadata', {}).get('uid'),
            name=event['object'].get('metadata', {}).get('name'),
            namespace=event['object'].get('metadata', {}).get('namespace'),
        )
    if 'cause' in kwargs:
        cause = kwargs.get('cause')
        kwargs.update(
            event=cause.event,
            body=cause.body,
            diff=cause.diff,
            old=cause.old,
            new=cause.new,
            patch=cause.patch,
            logger=cause.logger,
            spec=cause.body.setdefault('spec', {}),
            meta=cause.body.setdefault('metadata', {}),
            status=cause.body.setdefault('status', {}),
            uid=cause.body.get('metadata', {}).get('uid'),
            name=cause.body.get('metadata', {}).get('name'),
            namespace=cause.body.get('metadata', {}).get('namespace'),
        )

    if is_async_fn(fn):
        result = await fn(*args, **kwargs)
    else:

        # Not that we want to use functools, but for executors kwargs, it is officially recommended:
        # https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.run_in_executor
        real_fn = functools.partial(fn, *args, **kwargs)

        # Copy the asyncio context from current thread to the handlr's thread.
        # It can be copied 2+ times if there are sub-sub-handlers (rare case).
        context = contextvars.copy_context()
        real_fn = functools.partial(context.run, real_fn)

        loop = asyncio.get_event_loop()
        result = await loop.run_in_executor(config.WorkersConfig.get_syn_executor(), real_fn)
    return result


def is_async_fn(fn):
    if fn is None:
        return None
    elif isinstance(fn, functools.partial):
        return is_async_fn(fn.func)
    elif hasattr(fn, '__wrapped__'):  # @functools.wraps()
        return is_async_fn(fn.__wrapped__)
    else:
        return asyncio.iscoroutinefunction(fn)



import pytest

from kopf.engines.logging import ObjectLogger

OBJ1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}}
REF1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}


@pytest.mark.parametrize('logfn, event_type', [
    ['info', "Normal"],
    ['warning', "Warning"],
    ['error', "Error"],
    ['critical', "Fatal"],
])
async def test_posting_normal_levels(caplog, logstream, logfn, event_type, event_queue, event_queue_loop):
    logger = ObjectLogger(body=OBJ1)
    logger_fn = getattr(logger, logfn)

    logger_fn("hello %s", "world")

    assert event_queue.qsize() == 1
    event1 = event_queue.get_nowait()
    assert event1.ref == REF1
    assert event1.type == event_type
    assert event1.reason == "Logging"
    assert event1.message == "hello world"
    assert caplog.messages == ["hello world"]


@pytest.mark.parametrize('logfn', [
    'debug',
])
async def test_skipping_hidden_levels(caplog, logstream, logfn, event_queue, event_queue_loop):
    logger = ObjectLogger(body=OBJ1)
    logger_fn = getattr(logger, logfn)

    logger_fn("hello %s", "world")
    logger.info("must be here")

    assert event_queue.qsize() == 1  # not 2!
    assert caplog.messages == ["hello world", "must be here"]


@pytest.mark.parametrize('logfn', [
    'debug',
    'info',
    'warning',
    'error',
    'critical',
])
async def test_skipping_when_local_with_all_levels(caplog, logstream, logfn, event_queue, event_queue_loop):
    logger = ObjectLogger(body=OBJ1)
    logger_fn = getattr(logger, logfn)

    logger_fn("hello %s", "world", local=True)
    logger.info("must be here")

    assert event_queue.qsize() == 1  # not 2!
    assert caplog.messages == ["hello world", "must be here"]



import pytest

from kopf.structs.dicts import ensure


def test_existing_key():
    d = {'abc': {'def': {'hij': 'val'}}}
    ensure(d, ['abc', 'def', 'hij'], 'new')
    assert d == {'abc': {'def': {'hij': 'new'}}}


def test_unexisting_key_in_existing_dict():
    d = {'abc': {'def': {}}}
    ensure(d, ['abc', 'def', 'hij'], 'new')
    assert d == {'abc': {'def': {'hij': 'new'}}}


def test_unexisting_key_in_unexisting_dict():
    d = {}
    ensure(d, ['abc', 'def', 'hij'], 'new')
    assert d == {'abc': {'def': {'hij': 'new'}}}


def test_toplevel_key():
    d = {'key': 'val'}
    ensure(d, ['key'], 'new')
    assert d == {'key': 'new'}


def test_nonmapping_key():
    d = {'key': 'val'}
    with pytest.raises(TypeError):
        ensure(d, ['key', 'sub'], 'new')


def test_empty_path():
    d = {}
    with pytest.raises(ValueError) as e:
        ensure(d, [], 'new')
    assert "Setting a root of a dict is impossible" in str(e.value)



import kopf


@kopf.on.create('zalando.org', 'v1', 'kopfexamples')
def create_fn(spec, **kwargs):

    for item in spec.get('items', []):

        @kopf.on.this(id=item)
        async def create_item_fn(item=item, **kwargs):
            print(f"=== Handling creation for {item}. ===")



import pytest

import kopf


@pytest.fixture(params=[
    dict(strict=True),
    dict(strict=False),
    dict(),
], ids=['strict', 'relaxed', 'default'])
def all_modes_kwargs(request):
    return request.param


@pytest.fixture(params=[
    dict(strict=True),
], ids=['strict'])
def strict_kwargs(request):
    return request.param


@pytest.fixture(params=[
    dict(strict=False),
    dict(),  # no kwargs, the function defaults are used.
], ids=['relaxed', 'default'])
def relaxed_kwargs(request):
    return request.param

#
# No matter what `strict` mode is, the pre-existing names are preserved.
#

def test_preserved_name_of_dict(all_modes_kwargs):
    obj = {'metadata': {'name': 'preexisting-name'}}

    kopf.harmonize_naming(obj, name='provided-name', **all_modes_kwargs)

    assert 'name' in obj['metadata']
    assert 'generateName' not in obj['metadata']
    assert obj['metadata']['name'] == 'preexisting-name'


def test_preserved_names_of_multiple_objects(all_modes_kwargs, multicls):
    obj1 = {'metadata': {'name': 'preexisting-name-1'}}
    obj2 = {'metadata': {'name': 'preexisting-name-2'}}
    objs = multicls([obj1, obj2])

    kopf.harmonize_naming(objs, name='provided-name', **all_modes_kwargs)

    assert 'name' in obj1['metadata']
    assert 'generateName' not in obj1['metadata']
    assert obj1['metadata']['name'] == 'preexisting-name-1'

    assert 'name' in obj2['metadata']
    assert 'generateName' not in obj2['metadata']
    assert obj2['metadata']['name'] == 'preexisting-name-2'

#
# In strict mode and with the absent names, the provided name is used.
#

def test_assigned_name_of_dict(strict_kwargs):
    obj = {}

    kopf.harmonize_naming(obj, name='provided-name', **strict_kwargs)

    assert 'name' in obj['metadata']
    assert 'generateName' not in obj['metadata']
    assert obj['metadata']['name'] == 'provided-name'


def test_assigned_names_of_multiple_objects(strict_kwargs, multicls):
    obj1 = {'metadata': {'name': 'preexisting-name-1'}}
    obj2 = {'metadata': {'name': 'preexisting-name-2'}}
    objs = multicls([obj1, obj2])

    kopf.harmonize_naming(objs, name='provided-name', **strict_kwargs)

    assert 'name' in obj1['metadata']
    assert 'generateName' not in obj1['metadata']
    assert obj1['metadata']['name'] == 'preexisting-name-1'

    assert 'name' in obj2['metadata']
    assert 'generateName' not in obj2['metadata']
    assert obj2['metadata']['name'] == 'preexisting-name-2'

#
# In relaxed mode, if the names are absent, they are auto-generated.
#

def test_prefixed_name_of_dict(relaxed_kwargs):
    obj = {}

    kopf.harmonize_naming(obj, name='provided-name', **relaxed_kwargs)

    assert 'name' not in obj['metadata']
    assert 'generateName' in obj['metadata']
    assert obj['metadata']['generateName'] == 'provided-name-'


def test_prefixed_names_of_multiple_objects(relaxed_kwargs, multicls):
    obj1 = {}
    obj2 = {}
    objs = multicls([obj1, obj2])

    kopf.harmonize_naming(objs, name='provided-name', **relaxed_kwargs)

    assert 'name' not in obj1['metadata']
    assert 'generateName' in obj1['metadata']
    assert obj1['metadata']['generateName'] == 'provided-name-'

    assert 'name' not in obj2['metadata']
    assert 'generateName' in obj2['metadata']
    assert obj2['metadata']['generateName'] == 'provided-name-'



def match(handler, body, changed_fields=None):
    return (
        (not handler.field or _matches_field(handler, changed_fields or [])) and
        (not handler.labels or _matches_labels(handler, body)) and
        (not handler.annotations or _matches_annotations(handler, body))
    )


def _matches_field(handler, changed_fields):
    return any(field[:len(handler.field)] == handler.field for field in changed_fields)


def _matches_labels(handler, body):
    return _matches_metadata(handler=handler, body=body, metadata_type='labels')


def _matches_annotations(handler, body):
    return _matches_metadata(handler=handler, body=body, metadata_type='annotations')


def _matches_metadata(handler, body, metadata_type):
    metadata = getattr(handler, metadata_type)
    object_metadata = body.get('metadata', {}).get(metadata_type, {})

    for key, value in metadata.items():
        if key not in object_metadata:
            return False
        elif value is not None and value != object_metadata[key]:
            return False
        else:
            continue

    return True



import json

import pytest

from kopf.structs.lastseen import LAST_SEEN_ANNOTATION
from kopf.structs.lastseen import compute_digest
from kopf.structs.lastseen import get_state_diffs
from kopf.structs.lastseen import has_state, get_state
from kopf.structs.lastseen import retreive_state, refresh_state, freeze_state


def test_annotation_is_fqdn():
    assert LAST_SEEN_ANNOTATION.startswith('kopf.zalando.org/')


@pytest.mark.parametrize('expected, body', [
    pytest.param(False, {}, id='no-metadata'),
    pytest.param(False, {'metadata': {}}, id='no-annotations'),
    pytest.param(False, {'metadata': {'annotations': {}}}, id='no-lastseen'),
    pytest.param(True, {'metadata': {'annotations': {LAST_SEEN_ANNOTATION: ''}}}, id='present'),
])
def test_has_state(expected, body):
    result = has_state(body=body)
    assert result == expected


def test_get_state_removes_resource_references():
    body = {'apiVersion': 'group/version', 'kind': 'Kind'}
    state = get_state(body=body)
    assert state == {}


@pytest.mark.parametrize('field', [
    'uid',
    'name',
    'namespace',
    'selfLink',
    'generation',
    'finalizers',
    'resourceVersion',
    'creationTimestamp',
    'deletionTimestamp',
    'any-unexpected-field',
])
def test_get_state_removes_system_fields_and_cleans_parents(field):
    body = {'metadata': {field: 'x'}}
    state = get_state(body=body)
    assert state == {}


@pytest.mark.parametrize('field', [
    'uid',
    'name',
    'namespace',
    'selfLink',
    'generation',
    'finalizers',
    'resourceVersion',
    'creationTimestamp',
    'deletionTimestamp',
    'any-unexpected-field',
])
def test_get_state_removes_system_fields_but_keeps_extra_fields(field):
    body = {'metadata': {field: 'x', 'other': 'y'}}
    state = get_state(body=body, extra_fields=['metadata.other'])
    assert state == {'metadata': {'other': 'y'}}


@pytest.mark.parametrize('annotation', [
    pytest.param(LAST_SEEN_ANNOTATION, id='kopf'),
    pytest.param('kubectl.kubernetes.io/last-applied-configuration', id='kubectl'),
])
def test_get_state_removes_garbage_annotations_and_cleans_parents(annotation):
    body = {'metadata': {'annotations': {annotation: 'x'}}}
    state = get_state(body=body)
    assert state == {}


@pytest.mark.parametrize('annotation', [
    pytest.param(LAST_SEEN_ANNOTATION, id='kopf'),
    pytest.param('kubectl.kubernetes.io/last-applied-configuration', id='kubectl'),
])
def test_get_state_removes_garbage_annotations_but_keeps_others(annotation):
    body = {'metadata': {'annotations': {annotation: 'x', 'other': 'y'}}}
    state = get_state(body=body)
    assert state == {'metadata': {'annotations': {'other': 'y'}}}


def test_get_state_removes_status_and_cleans_parents():
    body = {'status': {'kopf': {'progress': 'x', 'anything': 'y'}, 'other': 'z'}}
    state = get_state(body=body)
    assert state == {}


def test_get_state_removes_status_but_keeps_extra_fields():
    body = {'status': {'kopf': {'progress': 'x', 'anything': 'y'}, 'other': 'z'}}
    state = get_state(body=body, extra_fields=['status.other'])
    assert state == {'status': {'other': 'z'}}


def test_get_state_clones_body():
    body = {'spec': {'depth': {'field': 'x'}}}
    state = get_state(body=body)
    body['spec']['depth']['field'] = 'y'
    assert state is not body
    assert state['spec'] is not body['spec']
    assert state['spec']['depth'] is not body['spec']['depth']
    assert state['spec']['depth']['field'] == 'x'


def test_refresh_state_patches_when_absent():
    body = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(body)  # json formatting can vary across interpreters
    patch = {}
    refresh_state(body=body, patch=patch)
    assert patch['metadata']['annotations'][LAST_SEEN_ANNOTATION] == encoded


def test_refresh_state_patches_when_present_and_is_different():
    body = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(body)  # json formatting can vary across interpreters
    body['metadata'] = {'annotations': {LAST_SEEN_ANNOTATION: '{}'}}
    patch = {}
    refresh_state(body=body, patch=patch)
    assert patch['metadata']['annotations'][LAST_SEEN_ANNOTATION] == encoded


def test_refresh_state_ignores_when_present_and_is_the_same():
    body = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(body)  # json formatting can vary across interpreters
    body['metadata'] = {'annotations': {LAST_SEEN_ANNOTATION: encoded}}
    patch = {}
    refresh_state(body=body, patch=patch)
    assert not patch


def test_freeze_state_patches_when_absent():
    body = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(body)  # json formatting can vary across interpreters
    patch = {}
    freeze_state(body=body, patch=patch)
    assert patch['status']['kopf']['frozen-state'] == encoded


def test_freeze_state_ignores_when_present():
    body = {'spec': {'depth': {'field': 'x'}},
            'status': {'kopf': {'frozen-state': '{}'}}}
    patch = {}
    freeze_state(body=body, patch=patch)
    assert not patch


def test_retreive_state_when_present():
    data = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(data)  # json formatting can vary across interpreters
    body = {'metadata': {'annotations': {LAST_SEEN_ANNOTATION: encoded}}}
    state = retreive_state(body=body)
    assert state == data


def test_retreive_state_when_absent():
    body = {}
    state = retreive_state(body=body)
    assert state is None


def test_compute_digest():
    body1 = {'spec': {'depth': {'field': 'x'}}}
    body2 = {'spec': {'depth': {'field': 'x'}}}
    digest1 = compute_digest(body=body1)
    digest2 = compute_digest(body=body2)
    assert isinstance(digest1, (str, int))
    assert isinstance(digest2, (str, int))
    assert digest1  # evals to true
    assert digest2  # evals to true
    assert digest1 == digest2


def test_state_changed_detected():
    data = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(data)  # json formatting can vary across interpreters
    body = {'metadata': {'annotations': {LAST_SEEN_ANNOTATION: encoded}}}
    old, new, diff = get_state_diffs(body=body)
    assert diff


def test_state_change_ignored_with_garbage_annotations():
    data = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(data)  # json formatting can vary across interpreters
    body = {'metadata': {'annotations': {LAST_SEEN_ANNOTATION: encoded}},
            'spec': {'depth': {'field': 'x'}}}
    old, new, diff = get_state_diffs(body=body)
    assert not diff


def test_state_changed_ignored_with_system_fields():
    data = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(data)  # json formatting can vary across interpreters
    body = {'metadata': {'annotations': {LAST_SEEN_ANNOTATION: encoded},
                         'finalizers': ['x', 'y', 'z'],
                         'generation': 'x',
                         'resourceVersion': 'x',
                         'creationTimestamp': 'x',
                         'deletionTimestamp': 'x',
                         'any-unexpected-field': 'x',
                         'uid': 'uid',
                         },
            'status': {'kopf': {'progress': 'x', 'anything': 'y'},
                       'other': 'x'
                       },
            'spec': {'depth': {'field': 'x'}}}
    old, new, diff = get_state_diffs(body=body)
    assert not diff


# This is to ensure it is callable with proper signature.
# For actual tests of diffing, see `/tests/diffs/`.
def test_state_diff():
    data = {'spec': {'depth': {'field': 'x'}}}
    encoded = json.dumps(data)  # json formatting can vary across interpreters
    body = {'metadata': {'annotations': {LAST_SEEN_ANNOTATION: encoded}},
            'status': {'x': 'y'},
            'spec': {'depth': {'field': 'y'}}}
    old, new, diff = get_state_diffs(body=body, extra_fields=['status.x'])
    assert old == {'spec': {'depth': {'field': 'x'}}}
    assert new == {'spec': {'depth': {'field': 'y'}}, 'status': {'x': 'y'}}
    assert len(diff) == 2  # spec.depth.field & status.x, but the order is not known.



import kopf


@kopf.on.create('zalando.org', 'v1', 'kopfexamples')
def create_fn(spec, **kwargs):
    print(f"And here we are! Creating: {spec}")
    return {'message': 'hello world'}  # will be the new status



import glob
import os.path
import pathlib
import subprocess

import pytest

from kopf.clients.auth import login

root_dir = os.path.relpath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
examples = sorted(glob.glob(os.path.join(root_dir, 'examples/*/')))
assert examples  # if empty, it is just the detection failed
examples = [path for path in examples if not glob.glob((os.path.join(path, 'test*.py')))]


@pytest.fixture(params=examples)
def exampledir(request):
    return pathlib.Path(request.param)


@pytest.fixture(scope='session', autouse=True)
def autologin(request):
    if request.config.getoption('--with-e2e') or request.config.getoption('--only-e2e'):
        login()  # or anything like that; it is not a unit-under-test


@pytest.fixture()
def with_crd():
    subprocess.run("kubectl apply -f examples/crd.yaml", shell=True, check=True)


@pytest.fixture()
def with_peering():
    subprocess.run("kubectl apply -f peering.yaml", shell=True, check=True)


@pytest.fixture()
def no_crd():
    subprocess.run("kubectl delete customresourcedefinition kopfexamples.zalando.org", shell=True, check=True)


@pytest.fixture()
def no_peering():
    subprocess.run("kubectl delete customresourcedefinition kopfpeerings.zalando.org", shell=True, check=True)



import asyncio
import logging

import pytest

import kopf
from kopf.reactor.causation import CREATE, UPDATE, DELETE, GONE, FREE, NOOP, RESUME, ACQUIRE, RELEASE
from kopf.reactor.handling import custom_object_handler
from kopf.structs.finalizers import FINALIZER
from kopf.structs.lastseen import LAST_SEEN_ANNOTATION


async def test_acquire(registry, handlers, resource, cause_mock,
                   caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = ACQUIRE

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert k8s_mocked.asyncio_sleep.call_count == 0
    assert k8s_mocked.sleep_or_wait.call_count == 0
    assert k8s_mocked.patch_obj.call_count == 1
    assert event_queue.empty()

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert 'metadata' in patch
    assert 'finalizers' in patch['metadata']
    assert FINALIZER in patch['metadata']['finalizers']

    assert_logs([
        "Adding the finalizer",
        "Patching with",
    ])


async def test_create(registry, handlers, resource, cause_mock,
                      caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = CREATE

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert handlers.create_mock.call_count == 1
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert k8s_mocked.asyncio_sleep.call_count == 0
    assert k8s_mocked.sleep_or_wait.call_count == 0
    assert k8s_mocked.patch_obj.call_count == 1
    assert not event_queue.empty()

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert 'metadata' in patch
    assert 'annotations' in patch['metadata']
    assert LAST_SEEN_ANNOTATION in patch['metadata']['annotations']
    assert 'status' in patch
    assert 'kopf' in patch['status']
    assert 'progress' in patch['status']['kopf']
    assert patch['status']['kopf']['progress'] is None  # 1 out of 1 handlers done

    assert_logs([
        "Creation event:",
        "Invoking handler 'create_fn'",
        "Handler 'create_fn' succeeded",
        "All handlers succeeded",
        "Patching with",
    ])


async def test_update(registry, handlers, resource, cause_mock,
                      caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = UPDATE

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert handlers.update_mock.call_count == 1
    assert not handlers.delete_mock.called

    assert k8s_mocked.asyncio_sleep.call_count == 0
    assert k8s_mocked.sleep_or_wait.call_count == 0
    assert k8s_mocked.patch_obj.call_count == 1
    assert not event_queue.empty()

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert 'metadata' in patch
    assert 'annotations' in patch['metadata']
    assert LAST_SEEN_ANNOTATION in patch['metadata']['annotations']
    assert 'status' in patch
    assert 'kopf' in patch['status']
    assert 'progress' in patch['status']['kopf']
    assert patch['status']['kopf']['progress'] is None  # 1 out of 1 handlers done

    assert_logs([
        "Update event:",
        "Invoking handler 'update_fn'",
        "Handler 'update_fn' succeeded",
        "All handlers succeeded",
        "Patching with",
    ])


async def test_delete(registry, handlers, resource, cause_mock,
                      caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = DELETE

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert handlers.delete_mock.call_count == 1

    assert k8s_mocked.asyncio_sleep.call_count == 0
    assert k8s_mocked.sleep_or_wait.call_count == 0
    assert k8s_mocked.patch_obj.call_count == 1
    assert not event_queue.empty()

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert 'status' in patch
    assert 'kopf' in patch['status']
    assert 'progress' in patch['status']['kopf']
    assert patch['status']['kopf']['progress'] is None  # 1 out of 1 handlers done

    assert_logs([
        "Deletion event",
        "Invoking handler 'delete_fn'",
        "Handler 'delete_fn' succeeded",
        "All handlers succeeded",
        "Removing the finalizer",
        "Patching with",
    ])


async def test_release(registry, resource, handlers, cause_mock, caplog, k8s_mocked, assert_logs):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = RELEASE
    cause_mock.body.setdefault('metadata', {})['finalizers'] = [FINALIZER]

    # register handlers (no deletion handlers)
    registry.register_cause_handler(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        event=RESUME,
        fn=lambda **_: None,
        requires_finalizer=False,
    )

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert k8s_mocked.asyncio_sleep.call_count == 0
    assert k8s_mocked.sleep_or_wait.call_count == 0
    assert k8s_mocked.patch_obj.call_count == 1
    assert event_queue.empty()

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert 'metadata' in patch
    assert 'finalizers' in patch['metadata']
    assert [] == patch['metadata']['finalizers']

    assert_logs([
        "Removing the finalizer",
        "Patching with",
    ])


#
# Informational causes: just log, and do nothing else.
#

async def test_gone(registry, handlers, resource, cause_mock,
                    caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = GONE

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert not k8s_mocked.asyncio_sleep.called
    assert not k8s_mocked.sleep_or_wait.called
    assert not k8s_mocked.patch_obj.called
    assert event_queue.empty()

    assert_logs([
        "Deleted, really deleted",
    ])


async def test_free(registry, handlers, resource, cause_mock,
                    caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = FREE

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert not k8s_mocked.asyncio_sleep.called
    assert not k8s_mocked.sleep_or_wait.called
    assert not k8s_mocked.patch_obj.called
    assert event_queue.empty()

    assert_logs([
        "Deletion event, but we are done with it",
    ])


async def test_noop(registry, handlers, resource, cause_mock,
                    caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = NOOP

    event_queue = asyncio.Queue()
    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert not k8s_mocked.asyncio_sleep.called
    assert not k8s_mocked.sleep_or_wait.called
    assert not k8s_mocked.patch_obj.called
    assert event_queue.empty()

    assert_logs([
        "Something has changed, but we are not interested",
    ])



import kopf


def test_adding_to_dict():
    obj = {}

    kopf.label(obj, {'label-1': 'value-1', 'label-2': 'value-2'})

    assert 'metadata' in obj
    assert 'labels' in obj['metadata']
    assert isinstance(obj['metadata']['labels'], dict)
    assert len(obj['metadata']['labels']) == 2
    assert 'label-1' in obj['metadata']['labels']
    assert 'label-2' in obj['metadata']['labels']
    assert obj['metadata']['labels']['label-1'] == 'value-1'
    assert obj['metadata']['labels']['label-2'] == 'value-2'


def test_adding_to_multiple_objects(multicls):
    obj1 = {}
    obj2 = {}
    objs = multicls([obj1, obj2])

    kopf.label(objs, {'label-1': 'value-1', 'label-2': 'value-2'})

    assert isinstance(obj1['metadata']['labels'], dict)
    assert len(obj1['metadata']['labels']) == 2
    assert 'label-1' in obj1['metadata']['labels']
    assert 'label-2' in obj1['metadata']['labels']
    assert obj1['metadata']['labels']['label-1'] == 'value-1'
    assert obj1['metadata']['labels']['label-2'] == 'value-2'

    assert isinstance(obj2['metadata']['labels'], dict)
    assert len(obj2['metadata']['labels']) == 2
    assert 'label-1' in obj2['metadata']['labels']
    assert 'label-2' in obj2['metadata']['labels']
    assert obj2['metadata']['labels']['label-1'] == 'value-1'
    assert obj2['metadata']['labels']['label-2'] == 'value-2'


def test_forcing_true():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    kopf.label(obj, {'label': 'new-value'}, force=True)
    assert obj['metadata']['labels']['label'] == 'new-value'


def test_forcing_false():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    kopf.label(obj, {'label': 'new-value'}, force=False)
    assert obj['metadata']['labels']['label'] == 'old-value'


def test_forcing_default():
    obj = {'metadata': {'labels': {'label': 'old-value'}}}
    kopf.label(obj, {'label': 'new-value'})
    assert obj['metadata']['labels']['label'] == 'old-value'


def test_nested_with_forced_true():
    obj = {'metadata': {'labels': {'label': 'old-value'}},
           'spec': {'template': {}}}
    kopf.label(obj, {'label': 'new-value'}, nested=['spec.template'], force=True)
    assert obj['metadata']['labels']['label'] == 'new-value'
    assert obj['spec']['template']['metadata']['labels']['label'] == 'new-value'


def test_nested_with_forced_false():
    obj = {'metadata': {'labels': {'label': 'old-value'}},
           'spec': {'template': {}}}
    kopf.label(obj, {'label': 'new-value'}, nested=['spec.template'], force=False)
    assert obj['metadata']['labels']['label'] == 'old-value'
    assert obj['spec']['template']['metadata']['labels']['label'] == 'new-value'



import asyncio
import concurrent.futures
import threading

import click.testing

from kopf import cli


class KopfRunner:
    """
    A context manager to run a Kopf-based operator in parallel with the tests.

    Usage::

        from kopf.testing import KopfRunner

        with KopfRunner(['run', '--verbose', 'examples/01-minimal/example.py']) as runner:
            # do something while the operator is running.
            time.sleep(3)

        assert runner.exit_code == 0
        assert runner.exception is None
        assert 'And here we are!' in runner.stdout

    All the args & kwargs are passed directly to Click's invocation method.
    See: `click.testing.CliRunner`.
    All properties proxy directly to Click's `click.testing.Result` object
    when it is available (i.e. after the context manager exits).

    CLI commands have to be invoked in parallel threads, never in processes:

    First, with multiprocessing, they are unable to pickle and pass
    exceptions (specifically, their traceback objects)
    from a child thread (Kopf's CLI) to the parent thread (pytest).

    Second, mocking works within one process (all threads),
    but not across processes --- the mock's calls (counts, arrgs) are lost.
    """

    def __init__(self, *args, reraise=True, timeout=None, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self.reraise = reraise
        self.timeout = timeout
        self._loop = asyncio.new_event_loop()
        self._ready = threading.Event()  # NB: not asyncio.Event!
        self._thread = threading.Thread(target=self._target)
        self._future = concurrent.futures.Future()

    def __enter__(self):
        self._thread.start()
        self._ready.wait()  # should be nanosecond-fast
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        # A coroutine that is injected into the loop to cancel everything in it.
        # Cancellations are caught in `run`, so that it exits gracefully.
        # TODO: also cancel/stop the streaming API calls in the thread executor.
        async def shutdown():
            current_task = asyncio.current_task()
            tasks = [task for task in asyncio.all_tasks() if task is not current_task]
            for task in tasks:
                task.cancel()

        # When the `with` block ends, shut down the parallel thread & loop
        # by cancelling all the tasks. Do not wait for the tasks to finish,
        # but instead wait for the thread+loop (CLI command) to finish.
        if self._loop.is_running():
            asyncio.run_coroutine_threadsafe(shutdown(), self._loop)
        self._thread.join(timeout=self.timeout)

        # If the thread is not finished, it is a bigger problem than exceptions.
        if self._thread.is_alive():
            raise Exception("The operator didn't stop, still running.")

        # Re-raise the exceptions of the threading & invocation logic.
        if self._future.exception() is not None:
            if exc_val is None:
                raise self._future.exception()
            else:
                raise self._future.exception() from exc_val
        if self._future.result().exception is not None and self.reraise:
            if exc_val is None:
                raise self._future.result().exception
            else:
                raise self._future.result().exception from exc_val

    def _target(self):

        # Every thread must have its own loop. The parent thread (pytest)
        # needs to know when the loop is set up, to be able to shut it down.
        asyncio.set_event_loop(self._loop)
        self._ready.set()

        # Execute the requested CLI command in the thread & thread's loop.
        # Remember the result & exception for re-raising in the parent thread.
        try:
            runner = click.testing.CliRunner()
            result = runner.invoke(cli.main, *self.args, **self.kwargs)
        except BaseException as e:
            self._future.set_exception(e)
        else:
            self._future.set_result(result)

    @property
    def future(self):
        return self._future

    @property
    def output(self):
        return self.future.result().output

    @property
    def stdout(self):
        return self.future.result().stdout

    @property
    def stdout_bytes(self):
        return self.future.result().stdout_bytes

    @property
    def stderr(self):
        return self.future.result().stderr

    @property
    def stderr_bytes(self):
        return self.future.result().stderr_bytes

    @property
    def exit_code(self):
        return self.future.result().exit_code

    @property
    def exception(self):
        return self.future.result().exception

    @property
    def exc_info(self):
        return self.future.result().exc_info



import collections.abc

import pytest

from kopf.clients.watching import streaming_watch


class PreventedActualCallError(Exception):
    pass


async def test_watching_over_cluster(resource, req_mock):
    req_mock.get.side_effect = PreventedActualCallError()

    itr = streaming_watch(
        resource=resource,
        namespace=None,
    )

    assert isinstance(itr, collections.abc.AsyncIterator)
    assert isinstance(itr, collections.abc.AsyncGenerator)

    with pytest.raises(PreventedActualCallError):
        async for _ in itr: pass  # fully deplete it

    assert req_mock.get.called
    assert req_mock.get.call_count == 1

    url = req_mock.get.call_args_list[0][1]['url']
    assert 'apis/zalando.org/v1/kopfexamples' in url
    assert 'namespaces/' not in url


async def test_watching_over_namespace(resource, req_mock):
    req_mock.get.side_effect = PreventedActualCallError()

    itr = streaming_watch(
        resource=resource,
        namespace='something',
    )

    assert isinstance(itr, collections.abc.AsyncIterator)
    assert isinstance(itr, collections.abc.AsyncGenerator)

    with pytest.raises(PreventedActualCallError):
        async for _ in itr: pass  # fully deplete it

    assert req_mock.get.called
    assert req_mock.get.call_count == 1

    url = req_mock.get.call_args_list[0][1]['url']
    assert 'apis/zalando.org/v1/namespaces/something/kopfexamples' in url



"""
Peer monitoring: knowing which other operators do run, and exchanging the basic signals with them.

The main use-case is to suppress all deployed operators when a developer starts a dev-/debug-mode
operator for the same cluster on their workstation -- to avoid any double-processing.

See also: `kopf freeze` & `kopf resume` CLI commands for the same purpose.

WARNING: There are **NO** per-object locks between the operators, so only one operator
should be functional for the cluster, i.e. only one with the highest priority running.
If the operator sees the violations of this constraint, it will print the warnings
pointing to another same-priority operator, but will continue to function.

The "signals" exchanged are only the keep-alive notifications from the operator being alive,
and detection of other operators hard termination (by timeout rather than by clear exit).

The peers monitoring covers both the in-cluster operators running,
and the dev-mode operators running in the dev workstations.

For this, special CRDs ``kind: ClusterKopfPeering`` & ``kind: KopfPeering``
should be registered in the cluster, and their ``status`` field is used
by all the operators to sync their keep-alive info.

The namespace-bound operators (e.g. `--namespace=`) report their individual
namespaces are part of the payload, can see all other cluster and namespaced
operators (even from the different namespaces), and behave accordingly.

The CRD is not applied automatically, so you have to deploy it yourself explicitly.
To disable the peers monitoring, use the `--standalone` CLI option.
"""

import asyncio
import datetime
import getpass
import logging
import os
import random
import socket
from typing import Iterable, Mapping, Optional, Union

import iso8601

from kopf.clients import fetching
from kopf.clients import patching
from kopf.reactor import registries
logger = logging.getLogger(__name__)

# The CRD info on the special sync-object.
CLUSTER_PEERING_RESOURCE = registries.Resource('zalando.org', 'v1', 'clusterkopfpeerings')
NAMESPACED_PEERING_RESOURCE = registries.Resource('zalando.org', 'v1', 'kopfpeerings')
LEGACY_PEERING_RESOURCE = registries.Resource('zalando.org', 'v1', 'kopfpeerings')
PEERING_DEFAULT_NAME = 'default'


# The class used to represent a peer in the parsed peers list (for convenience).
# The extra fields are for easier calculation when and if the peer is dead to the moment.
class Peer:

    def __init__(self,
                 id: str, *,
                 name: str,
                 priority: int = 0,
                 lastseen: Optional[str] = None,
                 lifetime: int = 60,
                 namespace: Optional[str] = None,
                 legacy: bool = False,
                 **kwargs):  # for the forward-compatibility with the new fields
        super().__init__()
        self.id = id
        self.name = name
        self.namespace = namespace
        self.priority = priority
        self.lifetime = (lifetime if isinstance(lifetime, datetime.timedelta) else
                         datetime.timedelta(seconds=int(lifetime)))
        self.lastseen = (lastseen if isinstance(lastseen, datetime.datetime) else
                         iso8601.parse_date(lastseen) if lastseen is not None else
                         datetime.datetime.utcnow())
        self.lastseen = self.lastseen.replace(tzinfo=None)  # only the naive utc -- for comparison
        self.deadline = self.lastseen + self.lifetime
        self.is_dead = self.deadline <= datetime.datetime.utcnow()
        self.legacy = legacy

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id}, namespace={self.namespace}, priority={self.priority}, lastseen={self.lastseen}, lifetime={self.lifetime})"

    @property
    def resource(self):
        return LEGACY_PEERING_RESOURCE if self.legacy else CLUSTER_PEERING_RESOURCE if self.namespace is None else NAMESPACED_PEERING_RESOURCE

    @classmethod
    def detect(cls,
               standalone: bool,
               namespace: Optional[str],
               name: Optional[str],
               **kwargs) -> Optional:

        if standalone:
            return None

        if name:
            if Peer._is_peering_exist(name, namespace=namespace):
                return cls(name=name, namespace=namespace, **kwargs)
            elif Peer._is_peering_legacy(name, namespace=namespace):
                return cls(name=name, namespace=namespace, legacy=True, **kwargs)
            else:
                raise Exception(f"The peering {name!r} was not found")

        if Peer._is_peering_exist(name=PEERING_DEFAULT_NAME, namespace=namespace):
            return cls(name=PEERING_DEFAULT_NAME, namespace=namespace, **kwargs)
        elif Peer._is_peering_legacy(name=PEERING_DEFAULT_NAME, namespace=namespace):
            return cls(name=PEERING_DEFAULT_NAME, namespace=namespace, legacy=True, **kwargs)

        logger.warning(f"Default peering object not found, falling back to the standalone mode.")
        return None

    def as_dict(self):
        # Only the non-calculated and non-identifying fields.
        return {
            'namespace': self.namespace,
            'priority': self.priority,
            'lastseen': self.lastseen.isoformat(),
            'lifetime': self.lifetime.total_seconds(),
        }

    def touch(self, *, lifetime: Optional[int] = None):
        self.lastseen = datetime.datetime.utcnow()
        self.lifetime = (self.lifetime if lifetime is None else
                         lifetime if isinstance(lifetime, datetime.timedelta) else
                         datetime.timedelta(seconds=int(lifetime)))
        self.deadline = self.lastseen + self.lifetime
        self.is_dead = self.deadline <= datetime.datetime.utcnow()

    async def keepalive(self):
        """
        Add a peer to the peers, and update its alive status.
        """
        self.touch()
        await apply_peers([self], name=self.name, namespace=self.namespace, legacy=self.legacy)

    async def disappear(self):
        """
        Remove a peer from the peers (gracefully).
        """
        self.touch(lifetime=0)
        await apply_peers([self], name=self.name, namespace=self.namespace, legacy=self.legacy)

    @staticmethod
    def _is_peering_exist(name: str, namespace: Optional[str]):
        resource = CLUSTER_PEERING_RESOURCE if namespace is None else NAMESPACED_PEERING_RESOURCE
        obj = fetching.read_obj(resource=resource, namespace=namespace, name=name, default=None)
        return obj is not None

    @staticmethod
    def _is_peering_legacy(name: str, namespace: Optional[str]):
        """
        Legacy mode for the peering: cluster-scoped KopfPeering (new mode: namespaced).

        .. deprecated:: 1.0

            This logic will be removed since 1.0.
            Deploy ``ClusterKopfPeering`` as per documentation, and use it normally.
        """
        crd = fetching.read_crd(resource=LEGACY_PEERING_RESOURCE, default=None)
        if crd is None:
            return False

        if str(crd.get('spec', {}).get('scope', '')).lower() != 'cluster':
            return False  # no legacy mode detected

        obj = fetching.read_obj(resource=LEGACY_PEERING_RESOURCE, name=name, default=None)
        return obj is not None


async def apply_peers(
        peers: Iterable[Peer],
        name: str,
        namespace: Union[None, str],
        legacy: bool = False,
):
    """
    Apply the changes in the peers to the sync-object.

    The dead peers are removed, the new or alive peers are stored.
    Note: this does NOT change their `lastseen` field, so do it explicitly with ``touch()``.
    """
    patch = {'status': {peer.id: None if peer.is_dead else peer.as_dict() for peer in peers}}
    resource = (LEGACY_PEERING_RESOURCE if legacy else
                CLUSTER_PEERING_RESOURCE if namespace is None else
                NAMESPACED_PEERING_RESOURCE)
    await patching.patch_obj(resource=resource, namespace=namespace, name=name, patch=patch)


async def peers_handler(
        *,
        event: Mapping,
        freeze: asyncio.Event,
        ourselves: Peer,
        autoclean: bool = True,
        replenished: asyncio.Event,
):
    """
    Handle a single update of the peers by us or by other operators.

    When an operator with a higher priority appears, switch to the freeze-mode.
    The these operators disappear or become presumably dead, resume the event handling.

    The freeze object is passed both to the peers handler to set/clear it,
    and to all the resource handlers to check its value when the events arrive
    (see `create_tasks` and `run` functions).
    """

    # Silently ignore the peering objects which are not ours to worry.
    body = event['object']
    name = body.get('metadata', {}).get('name', None)
    namespace = body.get('metadata', {}).get('namespace', None)
    if namespace != ourselves.namespace or name != ourselves.name:
        return

    # Find if we are still the highest priority operator.
    pairs = body.get('status', {}).items()
    peers = [Peer(id=opid, name=name, **opinfo) for opid, opinfo in pairs]
    dead_peers = [peer for peer in peers if peer.is_dead]
    prio_peers = [peer for peer in peers if not peer.is_dead and peer.priority > ourselves.priority]
    same_peers = [peer for peer in peers if not peer.is_dead and peer.priority == ourselves.priority and peer.id != ourselves.id]

    if autoclean and dead_peers:
        # NB: sync and blocking, but this is fine.
        await apply_peers(dead_peers, name=ourselves.name, namespace=ourselves.namespace, legacy=ourselves.legacy)

    if prio_peers:
        if not freeze.is_set():
            logger.info(f"Freezing operations in favour of {prio_peers}.")
            freeze.set()
    
    elif same_peers:
        logger.warning(f"Possibly conflicting operators with the same priority: {same_peers}.")
        logger.warning(f"Freezed all Operators: {peers}")
        freeze.set()
    elif freeze.is_set():
        logger.info(f"Resuming operations after the freeze. Conflicting operators with the same priority are gone")
        freeze.clear()


async def peers_keepalive(
        *,
        ourselves: Peer,
):
    """
    An ever-running coroutine to regularly send our own keep-alive status for the peers.
    """
    try:
        while True:
            logger.debug(f"Peering keep-alive update for {ourselves.id} (priority {ourselves.priority})")
            await ourselves.keepalive()

            # How often do we update. Keep limited to avoid k8s api flooding.
            # Should be slightly less than the lifetime, enough for a patch request to finish.
            await asyncio.sleep(max(1, int(ourselves.lifetime.total_seconds() - 10)))
    finally:
        try:
            await asyncio.shield(ourselves.disappear())
        except asyncio.CancelledError:
            # It is the cancellation of `keepalive()`, not of the shielded `disappear()`.
            pass
        except Exception:
            logger.exception(f"Couldn't remove self from the peering. Ignoring.")


def detect_own_id() -> str:
    """
    Detect or generate the id for ourselves, i.e. the execute operator.

    It is constructed easy to detect in which pod it is running
    (if in the cluster), or who runs the operator (if not in the cluster,
    i.e. in the dev-mode), and how long ago was it started.

    The pod id can be specified by::

        env:
        - name: POD_ID
          valueFrom:
            fieldRef:
              fieldPath: metadata.name

    Used in the `kopf.reactor.queueing` when the reactor starts,
    but is kept here, close to the rest of the peering logic.
    """

    pod = os.environ.get('POD_ID', None)
    if pod is not None:
        return pod

    user = getpass.getuser()
    host = socket.getfqdn()
    now = datetime.datetime.utcnow().isoformat()
    rnd = ''.join(random.choices('abcdefhijklmnopqrstuvwxyz0123456789', k=6))
    return f'{user}@{host}/{now}/{rnd}'
